from io import StringIO
import re

import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import (
    LinesReader,
)
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    read_csv,
    SeriesData,
)
from allotropy.parsers.utils.values import assert_not_none


class BeckmanEchoPlateReformatReader:
    SUPPORTED_EXTENSIONS = "csv"
    header: SeriesData
    sections: dict[str, pd.DataFrame]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = LinesReader.create(named_file_contents)

        # read header section, up to a data section delimiter
        header_lines = [
            line.strip() for line in reader.pop_until(r"^\[.+\]") if line.strip()
        ]
        if not header_lines:
            msg = "Cannot parse data from empty header."
            raise AllotropeConversionError(msg)

        # now read multiple sections, each starting with a section title: '[Section Title]', with header row and multiple data rows.
        sections = {}

        # we expect to have at least one section, and that the current line of the reader is at the title
        # (due to reader.pop_until(r"^\[.+\]") above)
        while not reader.is_empty():
            match = re.match(
                r"^\[(.+)\]",
                assert_not_none(reader.pop(), "Unexpected empty section"),
            )
            if match:
                title = str(
                    assert_not_none(
                        match, f"Cannot read title section: {reader.get()}"
                    ).groups()[0]
                )
                data_lines = list(
                    reader.pop_until_empty()
                )  # read all lines of section, including header line
                reader.drop_empty()  # sections are separated by empty lines

                sections[title] = read_csv(
                    StringIO("\n".join(data_lines)), sep=","
                ).replace(np.nan, None)
            else:
                break

        # read footer section, after last tabular data section
        footer_lines = list(reader.pop_until_empty())
        header_lines.extend(footer_lines)

        # parse header/footer to SeriesData
        with StringIO("\n".join(header_lines)) as csv_stream:
            raw_data = read_csv(
                csv_stream, header=None, sep=",", skipinitialspace=True, index_col=0
            )
        raw_data.index = raw_data.index.str.replace("*", "")
        header = df_to_series_data(raw_data.T.replace(np.nan, None))

        self.header = header
        self.sections = sections



from allotropy.calcdocs.extractor import Element, Extractor
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_structure import (
    Result,
)


class PerkinElmerEnvisionExtractor(Extractor[Result]):
    @classmethod
    def to_element(cls, result: Result) -> Element:
        return Element(
            uuid=result.uuid,
            data={
                "col": result.col,
                "row": result.row,
                "value": result.value,
            },
        )



from allotropy.allotrope.models.adm.binding_affinity_analyzer.wd._2024._12.binding_affinity_analyzer import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    Data as MapperData,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.cytiva_biacore_t200_control import constants
from allotropy.parsers.cytiva_biacore_t200_control.cytiva_biacore_t200_control_data_creator import (
    create_calculated_data,
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.cytiva_biacore_t200_control.cytiva_biacore_t200_control_decoder import (
    decode_data,
)
from allotropy.parsers.cytiva_biacore_t200_control.cytiva_biacore_t200_control_structure import (
    Data,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CytivaBiacoreT200ControlParser(VendorParser[MapperData, Model]):
    DISPLAY_NAME = constants.DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "blr"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> MapperData:
        data = Data.create(decode_data(named_file_contents))
        return MapperData(
            metadata=create_metadata(data, named_file_contents),
            measurement_groups=create_measurement_groups(data),
            calculated_data=create_calculated_data(data),
        )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import re

import numpy as np
import pandas as pd

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.thermo_skanit.constants import DEVICE_TYPE, SAMPLE_ROLE_MAPPINGS
from allotropy.parsers.utils.pandas import df_to_series_data, parse_header_row
from allotropy.parsers.utils.uuids import random_uuid_str

MEASUREMENT_TYPES = {
    MeasurementType.ULTRAVIOLET_ABSORBANCE: "Absorbance",
    MeasurementType.FLUORESCENCE: "Fluorescence",
    MeasurementType.LUMINESCENCE: "Luminescence",
}

SHEET_TABS = [
    "Session information",
    "Instrument information",
    "Layout definitions",
    "General information",
]

GENERAL_INFO_KEYS = [
    "software_name",
    "software_version",
]


@dataclass(frozen=True)
class ThermoSkanItMetadata:
    @staticmethod
    def create_metadata(
        file_path: str,
        instrument_info_df: pd.DataFrame | None,
        general_info_df: pd.DataFrame | None,
    ) -> Metadata:
        instrument_info_data = ThermoSkanItMetadata._get_instrument_data(
            instrument_info_df
        )
        general_info_data = ThermoSkanItMetadata._get_general_info_data(general_info_df)
        return Metadata(
            file_name=Path(file_path).name,
            unc_path=file_path,
            device_identifier=instrument_info_data["device_identifier"],
            model_number=instrument_info_data["model_number"],
            equipment_serial_number=instrument_info_data.get("equipment_serial_number"),
            software_name=general_info_data["software_name"],
            software_version=general_info_data["software_version"],
        )

    @staticmethod
    def _get_general_info_data(
        general_info_df: pd.DataFrame | None,
    ) -> dict[str, str | None]:
        if general_info_df is None:
            return dict.fromkeys(GENERAL_INFO_KEYS, None)

        general_info_data = df_to_series_data(parse_header_row(general_info_df.T))
        software_info = general_info_data.get(str, "Report generated with SW version")
        if not software_info:
            return dict.fromkeys(GENERAL_INFO_KEYS, NOT_APPLICABLE)
        software_name, software_version_txt = software_info.split(",", 1)
        match = re.search(r"\d+(?:\.\d+)*", software_version_txt)
        software_version = match.group() if match else None
        return {
            "software_name": software_name,
            "software_version": software_version,
        }

    @staticmethod
    def _get_instrument_data(
        instrument_info_df: pd.DataFrame | None,
    ) -> dict[str, str]:
        if instrument_info_df is None:
            return {
                "device_identifier": NOT_APPLICABLE,
                "model_number": NOT_APPLICABLE,
            }

        # Replace empty with "" so we can add label columns together
        instrument_info_df = instrument_info_df.fillna("")

        # The labels for data is spread across the first two columns for some reason, combine them as index.
        # NOTE: This is an assumption that may not be true for future files
        # Combine the first two columns into a Series
        combined_series = instrument_info_df.iloc[:, 0] + instrument_info_df.iloc[:, 1]
        # Convert the Series to an Index
        instrument_info_df.index = pd.Index(combined_series)

        # Read data from the last column, this is where values are found
        instrument_info_data = df_to_series_data(instrument_info_df.T, index=-1)

        lookups = {
            "device_identifier": ("Name", NOT_APPLICABLE),
            "model_number": ("Name", NOT_APPLICABLE),
            "equipment_serial_number": ("Serial number", None),
        }
        return {
            key: value
            for key, (label, default) in lookups.items()
            if (value := instrument_info_data.get(str, label, default)) is not None
        }


@dataclass(frozen=True)
class DataWell(Measurement):
    @staticmethod
    def create(
        well_location: str,
        value: float,
        sample_name: str,
        type_: MeasurementType,
        detector_wavelength: float | None,
        well_plate_identifier: str | None,
    ) -> Measurement:
        measurement_type_str = MEASUREMENT_TYPES[type_]
        return Measurement(
            type_=type_,
            identifier=random_uuid_str(),
            sample_identifier=sample_name,
            location_identifier=well_location,
            sample_role_type=DataWell.get_sample_role(sample_name),
            detector_wavelength_setting=detector_wavelength,
            device_type=DEVICE_TYPE,
            detection_type=measurement_type_str,
            well_plate_identifier=well_plate_identifier,
            absorbance=value
            if measurement_type_str
            == MEASUREMENT_TYPES[MeasurementType.ULTRAVIOLET_ABSORBANCE]
            else None,
            fluorescence=value
            if measurement_type_str == MEASUREMENT_TYPES[MeasurementType.FLUORESCENCE]
            else None,
            luminescence=value
            if measurement_type_str == MEASUREMENT_TYPES[MeasurementType.LUMINESCENCE]
            else None,
        )

    @staticmethod
    def get_sample_role(sample_name: str) -> SampleRoleType | None:
        stripped_text = re.sub(r"\d+", "", sample_name)
        try:
            return SAMPLE_ROLE_MAPPINGS[stripped_text]
        except KeyError as _:
            return None


@dataclass(frozen=True)
class ThermoSkanItMeasurementGroups:
    @staticmethod
    def create(
        sheet_df: pd.DataFrame,
        type_: MeasurementType,
        layout_definitions_df: pd.DataFrame | None,
        session_info_df: pd.DataFrame | None,
    ) -> list[MeasurementGroup]:
        (
            data_df,
            name_df,
            wavelength,
            well_plate_identifier,
        ) = ThermoSkanItMeasurementGroups.identify_data_and_sample_dfs(sheet_df)
        data_df.dropna(how="all", inplace=True)
        plate_well_count = None
        if layout_definitions_df is not None:
            plate_well_count = ThermoSkanItMeasurementGroups.get_plate_well_count(
                layout_definitions_df
            )
        if not plate_well_count:
            plate_well_count = data_df.size

        session_name = exec_time = None
        if session_info_df is not None:
            session_info_data = df_to_series_data(parse_header_row(session_info_df.T))
            session_name = session_info_data.get(str, "Session notes")
            exec_time = session_info_data.get(str, "Execution time")

        if not exec_time:
            exec_time = sheet_df.iloc[1].iloc[0]

        if not exec_time:
            msg = "Execution time not found"
            raise AllotropyParserError(msg)

        if not session_name:
            experiment = sheet_df.iloc[0].iloc[0]
            session_name = experiment.replace(".skax", "") if experiment else None

        meas_groups = []
        # Stack the DataFrame, creating a MultiIndex
        stacked = data_df.stack()

        # Iterate through the MultiIndex series and unpack it correctly
        for well_letter, well_column in stacked.index:
            if not name_df.empty:
                sample_name = name_df.loc[well_letter, well_column]
            else:
                sample_name = f"{well_plate_identifier}_{well_letter}{well_column}"
            well = DataWell.create(
                well_location=well_letter + str(well_column),
                value=stacked.loc[well_letter, well_column],
                sample_name=sample_name,
                detector_wavelength=wavelength,
                type_=type_,
                well_plate_identifier=well_plate_identifier,
            )
            meas_groups.append(
                MeasurementGroup(
                    measurements=[well],
                    plate_well_count=plate_well_count,
                    measurement_time=exec_time,
                    experimental_data_identifier=session_name,
                )
            )
        return meas_groups

    @staticmethod
    def get_plate_well_count(layout_definitions_df: pd.DataFrame) -> int | None:
        # Find row containing "Plate template"
        plate_row = layout_definitions_df[
            layout_definitions_df.iloc[:, 0].str.contains(
                "Plate template", case=False, na=False
            )
        ]

        if plate_row.empty:
            return None

        # Combine all non-empty values in the row
        combined_value = plate_row.iloc[0].dropna().str.cat(sep=" ")

        # Extract first number found
        if match := re.search(r"\d+", combined_value):
            return int(match.group())

        return None

    @staticmethod
    def _set_headers(df: pd.DataFrame) -> pd.DataFrame:
        # Set the first column (well letters) as the index
        df.set_index(df.columns[0], inplace=True)
        # Set the first row (well numbers) as the columns
        df = parse_header_row(df)
        # Cast row numbers to int (float first to handle decimals, e.g. 1.0)
        df.columns = df.columns.astype(float).astype(int)
        return df

    @staticmethod
    def identify_data_and_sample_dfs(
        absorbance_sheet_df: pd.DataFrame,
    ) -> tuple[pd.DataFrame, pd.DataFrame, float | None, str | None]:
        # Initialize variables
        start_reading_data = False
        start_reading_sample = False
        data_between_abs_and_blank = []
        data_between_sample_and_blank = []
        wavelength = None
        well_plate_identifier = None

        # Iterate through each row
        for _, row in absorbance_sheet_df.iterrows():
            if pd.notna(row.iloc[0]) and "Wavelength" in row.iloc[0]:
                match = re.search(r"Wavelength:\s*(\d{1,3})\s*nm", row.iloc[0])
                if match and (val := float(match.group(1))) != 0:
                    wavelength = val
            if pd.notna(row.iloc[0]) and "Plate" in row.iloc[0]:
                match = re.search(r"Plate\s*(\d)", row.iloc[0])
                if match:
                    well_plate_identifier = match.group(0)
            if "Abs" in row.values.astype(str) or "RLU" in row.values.astype(str):
                start_reading_data = True
            if "Sample" in row.values.astype(str):
                start_reading_sample = True

            if start_reading_sample:
                data_between_sample_and_blank.append(row)
            elif start_reading_data:
                data_between_abs_and_blank.append(row)

        # Create DataFrames with the collected data
        df_between_abs_and_blank = ThermoSkanItMeasurementGroups._set_headers(
            pd.DataFrame(data_between_abs_and_blank)
        )
        if data_between_sample_and_blank:
            df_between_sample_and_blank = ThermoSkanItMeasurementGroups._set_headers(
                pd.DataFrame(data_between_sample_and_blank)
            )
        else:
            df_between_sample_and_blank = pd.DataFrame()

        return (
            df_between_abs_and_blank,
            df_between_sample_and_blank,
            wavelength,
            well_plate_identifier,
        )


@dataclass(frozen=True)
class DataThermoSkanIt(Data):
    @staticmethod
    def _clean_dataframe(df: pd.DataFrame) -> pd.DataFrame:
        # Prevent pandas from silently downcasting values, to prevent future incompatibility.
        pd.set_option("future.no_silent_downcasting", True)  # noqa: FBT003
        df = df.replace(r"^\s*$", np.nan, regex=True)
        df = df.dropna(axis="index", how="all")
        df = df.dropna(axis="columns", how="all")
        return df

    @staticmethod
    def create(sheet_data: dict[str, pd.DataFrame], file_path: str) -> Data:
        measurement_df = _type = None
        clean_data = {
            key: DataThermoSkanIt._clean_dataframe(sheet_data[key])
            for key in sheet_data.keys()
            if key in SHEET_TABS
        }
        for measurement_type, measurement_type_key in MEASUREMENT_TYPES.items():
            for sheet_name, df in sheet_data.items():
                if measurement_type_key in sheet_name:
                    measurement_df = df
                    _type = measurement_type
                    break

        if measurement_df is None or measurement_df.empty or _type is None:
            msg = "Unsupported or missing measurement type"
            raise AllotropyParserError(msg)
        metadata = ThermoSkanItMetadata.create_metadata(
            instrument_info_df=clean_data.get("Instrument information"),
            general_info_df=clean_data.get("General information"),
            file_path=file_path,
        )
        measurement_groups = ThermoSkanItMeasurementGroups.create(
            sheet_df=measurement_df,
            layout_definitions_df=clean_data.get("Layout definitions"),
            session_info_df=clean_data.get("Session information"),
            type_=_type,
        )

        return Data(
            metadata=metadata,
            measurement_groups=measurement_groups,
        )



from __future__ import annotations

from collections.abc import Iterator
from io import StringIO
from re import search
from typing import Any, Literal

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.encoding import decode
from allotropy.parsers.utils.pandas import read_csv
from allotropy.parsers.utils.values import assert_not_none

EMPTY_STR_PATTERN = r"^\s*$"
# Use to match empty line in a CSV file
EMPTY_CSV_LINE = r"^,*$"
# Use to match empty string or empty CSV line
EMPTY_STR_OR_CSV_LINE = r"^[\s,]*$"


def read_to_lines(named_file_contents: NamedFileContents) -> list[str]:
    raw_contents = decode(named_file_contents.contents, named_file_contents.encoding)
    contents = raw_contents.replace("\r\n", "\n")
    return contents.split("\n")


class LinesReader:
    lines: list[str]
    current_line: int

    @staticmethod
    def create(named_file_contents: NamedFileContents) -> LinesReader:
        return LinesReader(read_to_lines(named_file_contents))

    def __init__(self, lines: list[str]) -> None:
        self.lines = lines
        self.current_line = 0

    def line_exists(self, line: int) -> bool:
        return 0 <= line < len(self.lines)

    def current_line_exists(self) -> bool:
        return self.line_exists(self.current_line)

    def get_line(self, line: int) -> str | None:
        return self.lines[line] if self.line_exists(line) else None

    def get(self) -> str | None:
        return self.lines[self.current_line] if self.current_line_exists() else None

    def match(self, match_pat: str) -> bool:
        line = self.get()
        return False if line is None else bool(search(match_pat, line))

    def is_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> bool:
        return self.match(empty_pat)

    def pop(self) -> str | None:
        line = self.get()
        if line is not None:
            self.current_line += 1
        return line

    def pop_if_match(self, match_pat: str) -> str | None:
        return self.pop() if self.match(match_pat) else None

    def pop_data(self) -> str | None:
        self.drop_empty()
        return self.pop()

    def drop_until(self, match_pat: str) -> str | None:
        while self.current_line_exists() and not self.match(match_pat):
            self.pop()
        return self.get()

    def drop_until_inclusive(self, match_pat: str) -> str | None:
        self.drop_until(match_pat)
        return self.pop()

    def drop_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> str | None:
        while self.current_line_exists() and self.is_empty(empty_pat):
            self.pop()
        return self.get()

    def drop_until_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> str | None:
        while self.current_line_exists() and not self.is_empty(empty_pat):
            self.pop()
        return self.get()

    def drop_until_empty_inclusive(
        self, empty_pat: str = EMPTY_STR_PATTERN
    ) -> str | None:
        self.drop_until_empty(empty_pat)
        return self.pop()

    def pop_while(self, match_pat: str) -> Iterator[str]:
        while self.current_line_exists() and self.match(match_pat):
            line = self.pop()
            if line is not None:
                yield line

    def pop_until(self, match_pat: str) -> Iterator[str]:
        while self.current_line_exists() and not self.match(match_pat):
            line = self.pop()
            if line is not None:
                yield line

    def pop_until_inclusive(self, match_pat: str) -> Iterator[str]:
        yield from self.pop_until(match_pat)
        if self.current_line_exists():
            yield assert_not_none(self.pop())

    def pop_until_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> Iterator[str]:
        while self.current_line_exists() and not self.is_empty(empty_pat):
            line = self.pop()
            if line is not None:
                yield line


class SectionLinesReader(LinesReader):
    @staticmethod
    def create(named_file_contents: NamedFileContents) -> SectionLinesReader:
        return SectionLinesReader(read_to_lines(named_file_contents))

    def iter_sections(self, pattern: str) -> Iterator[LinesReader]:
        self.drop_until(pattern)
        while True:
            if (initial_line := self.pop()) is None:
                break
            lines = [initial_line, *self.pop_until(pattern)]
            yield LinesReader(lines)


class InvertedLinesReader(LinesReader):
    def __init__(self, lines: list[str]) -> None:
        self.lines = lines
        self.current_line = len(lines) - 1

    def pop(self) -> str | None:
        line = self.get()
        if line is not None:
            self.current_line -= 1
        return line


class CsvReader(LinesReader):
    @staticmethod
    def create(named_file_contents: NamedFileContents) -> CsvReader:
        return CsvReader(read_to_lines(named_file_contents))

    def pop_csv_block_as_lines(self, empty_pat: str = EMPTY_STR_PATTERN) -> list[str]:
        self.drop_empty(empty_pat)
        lines = list(self.pop_until_empty(empty_pat))
        self.drop_empty(empty_pat)
        return lines

    def pop_csv_block_as_df(
        self,
        empty_pat: str = EMPTY_STR_PATTERN,
        header: int | Literal["infer"] | None = None,
        **kwargs: Any,
    ) -> pd.DataFrame | None:
        if lines := self.pop_csv_block_as_lines(empty_pat):
            return read_csv(
                StringIO("\n".join(lines)),
                # Prevent pandas from rounding decimal values, at the cost of some speed.
                float_precision="round_trip",
                header=header,
                **kwargs,
            )
        return None

    def drop_sections(self, match_pat: str) -> None:
        self.drop_empty()
        while self.match(match_pat):
            self.drop_until_empty()
            self.drop_empty()

    def pop_as_series(self, sep: str = " ") -> pd.Series[str] | None:
        line = self.pop()
        return None if line is None else pd.Series(line.split(sep))

    def lines_as_df(
        self,
        lines: list[str],
        header: int | Literal["infer"] | None = None,
        **kwargs: Any,
    ) -> pd.DataFrame | None:
        if lines:
            return read_csv(
                StringIO("\n".join(lines)),
                dtype=None,
                # Prevent pandas from rounding decimal values, at the cost of some speed.
                float_precision="round_trip",
                header=header,
                **kwargs,
            )
        return None



from __future__ import annotations

from collections import defaultdict

from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.view import View


class MeanView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="mean", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            if mean := element.get_float_or_none("mean"):
                items[str(mean)].append(element)
        return dict(items)


class SumView(View):
    def __init__(self) -> None:
        super().__init__(name="sum")

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            if sum_ := element.get_float_or_none("sum"):
                items[str(sum_)].append(element)
        return dict(items)



from collections.abc import Iterator

from allotropy.calcdocs.appbio_quantstudio_designandanalysis.config import (
    CalculatedDataConfigWithOptional,
)
from allotropy.calcdocs.appbio_quantstudio_designandanalysis.extractor import (
    AppbioQuantstudioDAExtractor,
)
from allotropy.calcdocs.appbio_quantstudio_designandanalysis.views import (
    SampleView,
    TargetRoleView,
    TargetView,
    UuidView,
)
from allotropy.calcdocs.config import (
    CalcDocsConfig,
    CalculatedDataConfig,
    MeasurementConfig,
)
from allotropy.calcdocs.view import ViewData
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    WellItem,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)


def ctr() -> MeasurementConfig:
    return MeasurementConfig(
        name="cycle threshold result",
        value="cycle_threshold_result",
        required=True,
    )


def norm_reporter_result() -> MeasurementConfig:
    return MeasurementConfig(
        name="normalized reporter result",
        value="normalized_reporter_result",
    )


def amplification_score(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="amplification score",
        value="amp_score",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def cq_confidence(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="cq confidence",
        value="cq_conf",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def y_intercept(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="y intercept",
        value="y_intercept",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def slope(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="slope",
        value="slope",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def ct_mean(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="ct mean",
        value="ct_mean",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def ct_sd(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="ct sd",
        value="ct_sd",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def ct_se(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="ct se",
        value="ct_se",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def r_squared(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="r^2",
        value="r_squared",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def efficiency(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="efficiency",
        value="efficiency",
        view_data=view_data,
        source_configs=(ctr(),),
    )


def rn_mean(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="rn mean",
        value="rn_mean",
        view_data=view_data,
        source_configs=(norm_reporter_result(),),
    )


def rn_sd(view_data: ViewData) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="rn sd",
        value="rn_sd",
        view_data=view_data,
        source_configs=(norm_reporter_result(),),
    )


def quantity(
    view_data: ViewData,
    y_intercept_conf: CalculatedDataConfig | None = None,
    slope_conf: CalculatedDataConfig | None = None,
) -> CalculatedDataConfig:
    ctr_conf = ctr()
    return CalculatedDataConfig(
        name="quantity",
        value="quantity",
        view_data=view_data,
        source_configs=tuple(
            config for config in [ctr_conf, y_intercept_conf, slope_conf] if config
        ),
    )


def quantity_mean(
    view_data: ViewData,
    quantity_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="quantity mean",
        value="quantity_mean",
        view_data=view_data,
        source_configs=(quantity_conf,),
    )


def quantity_sd(
    view_data: ViewData,
    quantity_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="quantity sd",
        value="quantity_sd",
        view_data=view_data,
        source_configs=(quantity_conf,),
    )


def delta_ct_sd(
    view_data: ViewData,
    ct_sd_conf: CalculatedDataConfig,
    ref_ct_sd_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="delta equivalent ct sd",
        value="delta_ct_sd",
        view_data=view_data,
        source_configs=(ct_sd_conf, ref_ct_sd_conf),
    )


def delta_ct_se(
    view_data: ViewData,
    ct_sd_conf: CalculatedDataConfig,
    ref_ct_sd_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="delta equivalent ct se",
        value="delta_ct_se",
        view_data=view_data,
        source_configs=(ct_sd_conf, ref_ct_sd_conf),
    )


def relative_rq(
    view_data: ViewData,
    quantity_mean_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="relative rq",
        value="rq",
        view_data=view_data,
        source_configs=(quantity_mean_conf,),
    )


def relative_rq_min(
    view_data: ViewData,
    rq_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="relative rq min",
        value="rq_min",
        view_data=view_data,
        source_configs=(rq_conf,),
    )


def relative_rq_max(
    view_data: ViewData,
    rq_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="relative rq max",
        value="rq_max",
        view_data=view_data,
        source_configs=(rq_conf,),
    )


def eq_ct_mean(
    view_data: ViewData,
    ct_mean_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="equivalent ct mean",
        value="eq_ct_mean",
        view_data=view_data,
        source_configs=(ct_mean_conf,),
    )


def adj_eq_ct_mean(
    view_data: ViewData,
    eq_ct_mean_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfigWithOptional(
        name="adjusted equivalent ct mean",
        value="adj_eq_ct_mean",
        view_data=view_data,
        optional=True,
        source_configs=(eq_ct_mean_conf,),
    )


def delta_ct(
    view_data: ViewData,
    adj_eq_ct_mean_conf: CalculatedDataConfig,
    ref_adj_eq_ct_mean_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="delta equivalent ct mean",
        value="delta_ct_mean",
        view_data=view_data,
        source_configs=(adj_eq_ct_mean_conf, ref_adj_eq_ct_mean_conf),
    )


def delta_delta_ct(
    view_data: ViewData,
    delta_ct_conf: CalculatedDataConfig,
    ref_delta_ct_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="delta delta equivalent ct",
        value="delta_delta_ct",
        view_data=view_data,
        source_configs=(delta_ct_conf, ref_delta_ct_conf),
    )


def rq(
    view_data: ViewData,
    delta_delta_ct_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="rq",
        value="rq",
        view_data=view_data,
        source_configs=(delta_delta_ct_conf,),
    )


def rq_min(
    view_data: ViewData,
    rq_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="rq min",
        value="rq_min",
        view_data=view_data,
        source_configs=(rq_conf,),
    )


def rq_max(
    view_data: ViewData,
    rq_conf: CalculatedDataConfig,
) -> CalculatedDataConfig:
    return CalculatedDataConfig(
        name="rq max",
        value="rq_max",
        view_data=view_data,
        source_configs=(rq_conf,),
    )


def iter_standard_curve_calc_docs(
    well_items: list[WellItem],
) -> Iterator[CalculatedDocument]:
    # Y-intercept, Slope, Quantity, Amp score, Cq confidence,
    # Quantity Mean, Quantity SD, Ct Mean, Ct SD
    # R^2, Efficiency
    elements = AppbioQuantstudioDAExtractor.get_elements(well_items)

    sid_tdna_view_data = SampleView(sub_view=TargetView()).apply(elements)
    sid_tdna_uuid_view_data = SampleView(
        sub_view=TargetView(sub_view=UuidView())
    ).apply(elements)
    tdna_view_data = TargetRoleView().apply(elements)

    quantity_conf = quantity(
        sid_tdna_uuid_view_data,
        y_intercept(tdna_view_data),
        slope(tdna_view_data),
    )

    configs = CalcDocsConfig(
        [
            quantity_conf,
            amplification_score(sid_tdna_uuid_view_data),
            cq_confidence(sid_tdna_uuid_view_data),
            quantity_mean(sid_tdna_view_data, quantity_conf),
            quantity_sd(sid_tdna_view_data, quantity_conf),
            ct_mean(sid_tdna_view_data),
            ct_sd(sid_tdna_view_data),
            y_intercept(tdna_view_data),
            r_squared(tdna_view_data),
            slope(tdna_view_data),
            efficiency(tdna_view_data),
        ]
    )

    for calc_doc in configs.construct():
        yield from calc_doc.iter_struct()


def iter_relative_standard_curve_calc_docs(
    well_items: list[WellItem],
    r_sample: str,
    r_target: str | None,
) -> Iterator[CalculatedDocument]:
    # Y-Intercept, Slope, Quantity, Amp score, Cq confidence,
    # Ct Mean, Ct SD, Ct SE,
    # Delta Ct SD, Delta Ct SE,
    # Relative RQ min, Relative RQ max, Relative RQ, Quantity Mean,
    # RQ min, RQ max, RQ, Delta Delta Ct, Delta Ct,
    # Adjusted equivalent Ct mean, equivalent Ct mean
    elements = AppbioQuantstudioDAExtractor.get_elements(well_items)

    sid_tdna_view_data = SampleView(sub_view=TargetView()).apply(elements)
    sid_ref_tdna_view_data = SampleView(
        reference=r_sample, sub_view=TargetView()
    ).apply(elements)
    sid_tdna_ref_view_data = SampleView(
        sub_view=TargetView(is_reference=True, reference=r_target)
    ).apply(elements)
    sid_tdna_blacklist_view_data = SampleView(
        sub_view=TargetView(blacklist=[r_target] if r_target is not None else None)
    ).apply(elements)
    sid_tdna_uuid_view_data = SampleView(
        sub_view=TargetView(sub_view=UuidView())
    ).apply(elements)
    tdna_view_data = TargetRoleView().apply(elements)

    quantity_conf = quantity(
        sid_tdna_uuid_view_data,
        y_intercept(tdna_view_data),
        slope(tdna_view_data),
    )

    configs = CalcDocsConfig(
        [
            quantity_conf,
            amplification_score(sid_tdna_uuid_view_data),
            cq_confidence(sid_tdna_uuid_view_data),
            ct_mean(sid_tdna_view_data),
            ct_sd(sid_tdna_view_data),
            delta_ct_sd(
                sid_tdna_view_data,
                ct_sd(sid_tdna_view_data),
                ct_sd(sid_tdna_ref_view_data),
            ),
            delta_ct_se(
                sid_tdna_view_data,
                ct_se(sid_tdna_view_data),
                ct_se(sid_tdna_ref_view_data),
            ),
            relative_rq_min(
                sid_tdna_view_data,
                relative_rq(
                    sid_tdna_view_data,
                    quantity_mean(sid_tdna_view_data, quantity_conf),
                ),
            ),
            relative_rq_max(
                sid_tdna_view_data,
                relative_rq(
                    sid_tdna_view_data,
                    quantity_mean(sid_tdna_view_data, quantity_conf),
                ),
            ),
            rq_min(
                sid_tdna_blacklist_view_data,
                rq(
                    sid_tdna_view_data,
                    delta_delta_ct(
                        sid_tdna_view_data,
                        delta_ct(
                            sid_tdna_view_data,
                            adj_eq_ct_mean(
                                sid_tdna_view_data,
                                eq_ct_mean(
                                    sid_tdna_view_data,
                                    ct_mean(sid_tdna_view_data),
                                ),
                            ),
                            adj_eq_ct_mean(
                                sid_tdna_ref_view_data,
                                eq_ct_mean(
                                    sid_tdna_view_data,
                                    ct_mean(sid_tdna_view_data),
                                ),
                            ),
                        ),
                        delta_ct(
                            sid_ref_tdna_view_data,
                            adj_eq_ct_mean(
                                sid_tdna_view_data,
                                eq_ct_mean(
                                    sid_tdna_view_data,
                                    ct_mean(sid_tdna_view_data),
                                ),
                            ),
                            adj_eq_ct_mean(
                                sid_tdna_ref_view_data,
                                eq_ct_mean(
                                    sid_tdna_view_data,
                                    ct_mean(sid_tdna_view_data),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            rq_max(
                sid_tdna_blacklist_view_data,
                rq(
                    sid_tdna_view_data,
                    delta_delta_ct(
                        sid_tdna_view_data,
                        delta_ct(
                            sid_tdna_view_data,
                            adj_eq_ct_mean(
                                sid_tdna_view_data,
                                eq_ct_mean(
                                    sid_tdna_view_data,
                                    ct_mean(sid_tdna_view_data),
                                ),
                            ),
                            adj_eq_ct_mean(
                                sid_tdna_ref_view_data,
                                eq_ct_mean(
                                    sid_tdna_view_data,
                                    ct_mean(sid_tdna_view_data),
                                ),
                            ),
                        ),
                        delta_ct(
                            sid_ref_tdna_view_data,
                            adj_eq_ct_mean(
                                sid_tdna_view_data,
                                eq_ct_mean(
                                    sid_tdna_view_data,
                                    ct_mean(sid_tdna_view_data),
                                ),
                            ),
                            adj_eq_ct_mean(
                                sid_tdna_ref_view_data,
                                eq_ct_mean(
                                    sid_tdna_view_data,
                                    ct_mean(sid_tdna_view_data),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        ]
    )

    for calc_doc in configs.construct():
        yield from calc_doc.iter_struct()


def iter_presence_absence_calc_docs(
    well_items: list[WellItem],
) -> Iterator[CalculatedDocument]:
    # Quantity, Amp score, Cq confidence, Rn Mean, Rn SD
    elements = AppbioQuantstudioDAExtractor.get_elements(well_items)

    sid_tdna_view_data = SampleView(sub_view=TargetView()).apply(elements)
    sid_tdna_uuid_view_data = SampleView(
        sub_view=TargetView(sub_view=UuidView())
    ).apply(elements)

    configs = CalcDocsConfig(
        [
            quantity(sid_tdna_uuid_view_data),
            amplification_score(sid_tdna_uuid_view_data),
            cq_confidence(sid_tdna_uuid_view_data),
            rn_mean(sid_tdna_view_data),
            rn_sd(sid_tdna_view_data),
        ]
    )

    for calc_doc in configs.construct():
        yield from calc_doc.iter_struct()


def iter_primary_analysis_calc_docs(
    well_items: list[WellItem],
) -> Iterator[CalculatedDocument]:
    # Ct Mean, Ct SD, Ct SE
    elements = AppbioQuantstudioDAExtractor.get_elements(well_items)

    sid_tdna_view_data = SampleView(sub_view=TargetView()).apply(elements)

    configs = CalcDocsConfig(
        [
            ct_mean(sid_tdna_view_data),
            ct_sd(sid_tdna_view_data),
            ct_se(sid_tdna_view_data),
        ]
    )

    for calc_doc in configs.construct():
        yield from calc_doc.iter_struct()



"""Maps the intermediate json to liquid chromatograhy mapper fields"""
import os
from typing import Any

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    DeviceDocument,
    Measurement,
    MeasurementGroup,
    Metadata,
    Peak,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_openlab_cds import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float_or_none,
)


def create_data_cube(
    label: str,
    dimension_value: list[float],
    measures_value: list[float],
    data_cube_component: DataCubeComponent,
) -> DataCube:
    return DataCube(
        label=label,
        structure_dimensions=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="retention time",
                unit="s",
            ),
        ],
        structure_measures=[data_cube_component],
        dimensions=[dimension_value],
        measures=[measures_value],
    )


def create_metadata(
    intermediate_structured_data: dict[str, Any], named_file_contents: NamedFileContents
) -> Metadata:
    return Metadata(
        asset_management_identifier=NOT_APPLICABLE,
        analyst=intermediate_structured_data["Metadata"]["Instrument"].get(
            "CreatedByUser"
        ),
        detection_type=intermediate_structured_data["Metadata"]["Instrument"].get(
            "Technique"
        ),
        model_number=intermediate_structured_data["Metadata"]["Instrument"].get("Name"),
        software_name=intermediate_structured_data["Metadata"]["Instrument"][
            "AcquisitionApplication"
        ]
        .get("AgilentApp", {})
        .get("Name"),
        file_name=os.path.basename(named_file_contents.original_file_path),
        data_system_instance_identifier=NOT_APPLICABLE,
        unc_path=named_file_contents.original_file_path,
        software_version=intermediate_structured_data["Metadata"]["Instrument"][
            "AcquisitionApplication"
        ]
        .get("AgilentApp", {})
        .get("Version"),
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        brand_name=(
            intermediate_structured_data["Metadata"]["Instrument"]["Name"]
        ).split()[1]
        if intermediate_structured_data["Metadata"]["Instrument"]["Name"]
        else None,
        device_identifier=intermediate_structured_data["Metadata"]["Instrument"].get(
            "@id"
        ),
        device_documents=[
            DeviceDocument(
                device_type=assert_not_none(module.get("Type")),
                device_identifier=module.get("@id"),
                product_manufacturer=module.get("Manufacturer"),
                model_number=module.get("PartNo"),
                equipment_serial_number=module.get("SerialNo"),
                firmware_version=module.get("FirmwareRevision"),
                device_custom_info={"device name": module["Name"]}
                if module["Name"]
                else None,
            )
            for module in intermediate_structured_data["Metadata"]["Instrument"][
                "Module"
            ]
            if module["Type"] not in ["ColumnCompartment", "Detector"]
        ],
    )


def create_peak(peak_structure: list[dict[str, Any]]) -> list[Peak]:
    return [
        Peak(
            identifier=peak["@id"],
            start=float(peak["BeginTime"]["@val"]) * 60
            if "Area" in peak and peak["BeginTime"].get("@val")
            else None,
            end=float(peak["EndTime"]["@val"]) * 60
            if "Area" in peak and peak["EndTime"].get("@val")
            else None,
            start_unit="s",
            end_unit="s",
            area=try_float_or_none(peak["Area"]["@val"]),
            area_unit="RFU.s"
            if "LU" in peak["Area"]["@unit"]
            else "mAU.s"
            if "Area" in peak and peak["Area"].get("@unit")
            else None,
            height=try_float_or_none(peak.get("Height", {}).get("@val"))
            if try_float_or_none(peak.get("Height", {}).get("@val")) is not None
            else None,
            height_unit="RFU"
            if "LU" in peak["Height"]["@unit"]
            else "mAU"
            if "Height" in peak and peak["Height"].get("@unit")
            else None,
            written_name=peak["Peak Metadata"].get("CompoundName"),
            retention_time=float(peak["RetentionTime"]["@val"]) * 60
            if peak.get("RetentionTime")
            and peak["RetentionTime"].get("@val") is not None
            else None,
            chromatographic_resolution=try_float_or_none(
                peak.get("Resolution_USP", {}).get("@val")
            ),
            width_at_half_height=float(peak.get("Width_50Perc", {}).get("@val")) * 60
            if peak.get("Width_50Perc") and peak["Width_50Perc"].get("@val") is not None
            else None,
            asymmetry_factor_measured_at_10_percent_height=try_float_or_none(
                peak.get("Width_50Perc", {}).get("@val")
            ),
            width_at_half_height_unit="s",
            peak_width_at_5_percent_of_height=float(
                peak.get("Width_5Perc", {}).get("@val")
            )
            * 60
            if peak.get("Width_5Perc") and peak["Width_5Perc"].get("@val") is not None
            else None,
            peak_width_at_10_percent_of_height=float(
                peak.get("Width_10Perc", {}).get("@val")
            )
            * 60
            if peak.get("Width_10Perc") and peak["Width_10Perc"].get("@val") is not None
            else None,
            peak_width_at_baseline=float(peak.get("WidthBase", {}).get("@val")) * 60
            if peak.get("WidthBase") and peak["WidthBase"].get("@val") is not None
            else None,
            number_of_theoretical_plates__chromatography_=try_float_or_none(
                peak.get("TheoreticalPlates_USP", {}).get("@val")
            ),
        )
        for peak in peak_structure
    ]


def create_measurements(
    intermediate_structured_data: dict[str, Any]
) -> list[Measurement]:
    column_comp_dict = next(
        (
            module
            for module in intermediate_structured_data["Metadata"]["Instrument"][
                "Module"
            ]
            if module["Name"] == "Column Comp."
        ),
        None,
    )
    return [
        Measurement(
            measurement_identifier=random_uuid_str(),
            processed_data_identifier=random_uuid_str(),
            measurement_time=intermediate_structured_data["Metadata"]["Instrument"].get(
                "CreationDate"
            ),
            sample_identifier=intermediate_structured_data["Result Data"][i][
                "Sample Data"
            ].get("SampleName"),
            sample_role_type=constants.SAMPLE_ROLE_TYPE.get(
                intermediate_structured_data["Result Data"][i]["Sample Data"][
                    "SampleSetup"
                ]
                .get("DAParam", {})
                .get("Type")
            ),
            sample_custom_info={
                "location identifier": intermediate_structured_data["Result Data"][i][
                    "Sample Data"
                ].get("Location"),
                "replicate": intermediate_structured_data["Result Data"][i][
                    "Sample Data"
                ].get("Replicate"),
                "sample amount": {
                    "value": try_float_or_none(
                        intermediate_structured_data["Result Data"][i]["Sample Data"][
                            "SampleSetup"
                        ]
                        .get("DAParam", {})
                        .get("SampleAmount", {})
                        .get("@val")
                    ),
                    "unit": intermediate_structured_data["Result Data"][i][
                        "Sample Data"
                    ]["SampleSetup"]["DAParam"]["SampleAmount"]["@unit"]
                    or "(unitless)",
                }
                if "SampleAmount"
                in intermediate_structured_data["Result Data"][i]["Sample Data"][
                    "SampleSetup"
                ]["DAParam"]
                else None,
                "barcode": intermediate_structured_data["Result Data"][i][
                    "Sample Data"
                ].get("Barcode"),
            },
            injection_identifier=assert_not_none(
                (
                    intermediate_structured_data["Result Data"][i]["Sample Data"][
                        "SampleMeasurement"
                    ]
                    .get("InjectionMeasData_ID", {})
                    .get("@id")
                )
                if isinstance(
                    intermediate_structured_data["Result Data"][i]["Sample Data"][
                        "SampleMeasurement"
                    ]["InjectionMeasData_ID"],
                    dict,
                )
                else (
                    intermediate_structured_data["Result Data"][i]["Sample Data"][
                        "SampleMeasurement"
                    ]["InjectionMeasData_ID"][
                        int(
                            intermediate_structured_data["Result Data"][i][
                                "Sample Data"
                            ]["Replicate"]
                        )
                        - 1
                    ].get(
                        "@id"
                    )
                )
            ),
            injection_time=assert_not_none(
                intermediate_structured_data["Result Data"][i]["Sample Data"].get(
                    "RunDateTime"
                )
            ),
            autosampler_injection_volume_setting=assert_not_none(
                float(
                    intermediate_structured_data["Result Data"][i]["Sample Data"].get(
                        "InjectionVolume"
                    )
                )
            ),
            injection_custom_info={
                "injection source": intermediate_structured_data["Result Data"][i][
                    "Sample Data"
                ].get("InjectionSource"),
                "acquisition method identifier": intermediate_structured_data[
                    "Result Data"
                ][i]["Sample Data"].get("AcquisitionMethod"),
                "injector position": intermediate_structured_data["Result Data"][i][
                    "Sample Data"
                ]["SampleSetup"]
                .get("AcqParam", {})
                .get("InjectorPosition"),
            },
            chromatography_serial_num=column_comp_dict.get("SerialNo")
            if column_comp_dict
            else None,
            chromatography_part_num=column_comp_dict.get("PartNo")
            if column_comp_dict
            else None,
            column_inner_diameter=try_float_or_none(
                intermediate_structured_data["Metadata"]["SeparationMedium"][0]["Type"][
                    "Column"
                ]
                .get("Diameter", {})
                .get("@val")
            ),
            chromatography_particle_size=try_float_or_none(
                intermediate_structured_data["Metadata"]["SeparationMedium"][0]["Type"][
                    "Column"
                ]
                .get("ParticleSize", {})
                .get("@val")
            ),
            chromatography_length=float(
                intermediate_structured_data["Metadata"]["SeparationMedium"][0]["Type"][
                    "Column"
                ]["Length"]["@val"]
            )
            / 10
            if "Length"
            in intermediate_structured_data["Metadata"]["SeparationMedium"][0]["Type"][
                "Column"
            ]
            else None,
            column_product_manufacturer=column_comp_dict.get("Manufacturer")
            if column_comp_dict
            else None,
            column_custom_info={
                "maximum temperature": {
                    "value": try_float_or_none(
                        intermediate_structured_data["Metadata"]["SeparationMedium"][0][
                            "Type"
                        ]["Column"]["MaxTemp"]["@val"]
                    ),
                    "unit": "degC"
                    if intermediate_structured_data["Metadata"]["SeparationMedium"][0][
                        "Type"
                    ]["Column"]["MaxTemp"]["@unit"]
                    == "°C"
                    else (
                        intermediate_structured_data["Metadata"]["SeparationMedium"][0][
                            "Type"
                        ]["Column"]["MaxTemp"]["@unit"]
                        or "(unitless)"
                    ),
                }
                if "MaxTemp"
                in intermediate_structured_data["Metadata"]["SeparationMedium"][0][
                    "Type"
                ]["Column"]
                else None,
                "maximum pressure": {
                    "value": try_float_or_none(
                        intermediate_structured_data["Metadata"]["SeparationMedium"][0][
                            "Type"
                        ]["Column"]["MaxPressure"]["@val"]
                    ),
                    "unit": intermediate_structured_data["Metadata"][
                        "SeparationMedium"
                    ][0]["Type"]["Column"]["MaxPressure"]["@unit"]
                    or "(unitless)",
                }
                if "MaxPressure"
                in intermediate_structured_data["Metadata"]["SeparationMedium"][0][
                    "Type"
                ]["Column"]
                else None,
                "firmware version": column_comp_dict.get("FirmwareRevision")
                if column_comp_dict
                else None,
            },
            device_control_docs=[
                DeviceControlDoc(
                    device_type="pump",
                    system_pressure_data_cube=create_data_cube(
                        label="pump pressure",
                        measures_value=intermediate_structured_data["Result Data"][i][
                            "Pressure"
                        ],
                        dimension_value=intermediate_structured_data["Result Data"][i][
                            "Time"
                        ],
                        data_cube_component=DataCubeComponent(
                            type_=FieldComponentDatatype.double,
                            concept="pump pressure",
                            unit="MPa",
                        ),
                    ),
                )
            ]
            if "Pressure"
            in intermediate_structured_data["Result Data"][i]["Metadata"]["signal"]
            else [
                DeviceControlDoc(
                    device_type=assert_not_none(module.get("Name")),
                    detection_type="Fluorescence"
                    if module.get("Name") == "FLD"
                    else "Absorbance",
                    device_identifier=module.get("@id"),
                    product_manufacturer=module.get("Manufacturer"),
                    model_number=module.get("PartNo"),
                    equipment_serial_number=module.get("SerialNo"),
                    firmware_version=module.get("FirmwareRevision"),
                    excitation_wavelength_setting=try_float_or_none(
                        intermediate_structured_data["Result Data"][i]["Metadata"][
                            "signal"
                        ]
                        .split("Ex=")[1]
                        .split(",")[0]
                    )
                    if module["Name"] == "FLD"
                    and "FLD"
                    in intermediate_structured_data["Result Data"][i]["Metadata"][
                        "signal"
                    ]
                    else None,
                    detector_bandwidth_setting=try_float_or_none(
                        intermediate_structured_data["Result Data"][i]["Metadata"][
                            "signal"
                        ]
                        .split(",")[2]
                        .split()[0]
                    )
                    if module["Name"] == "DAD"
                    and "DAD"
                    in intermediate_structured_data["Result Data"][i]["Metadata"][
                        "signal"
                    ]
                    else None,
                    detector_wavelength_setting=try_float_or_none(
                        intermediate_structured_data["Result Data"][i]["Metadata"][
                            "signal"
                        ]
                        .split("Sig=")[1]
                        .split(",")[0]
                    )
                    if "DAD" in module["Name"]
                    and "DAD"
                    in intermediate_structured_data["Result Data"][i]["Metadata"][
                        "signal"
                    ]
                    else try_float_or_none(
                        intermediate_structured_data["Result Data"][i]["Metadata"][
                            "signal"
                        ]
                        .split("Em=")[1]
                        .strip()
                    )
                    if module["Name"] == "FLD"
                    and "FLD"
                    in intermediate_structured_data["Result Data"][i]["Metadata"][
                        "signal"
                    ]
                    else None,
                )
                for module in intermediate_structured_data["Metadata"]["Instrument"][
                    "Module"
                ]
                if (
                    "DAD" in module["Name"]
                    and "DAD"
                    in intermediate_structured_data["Result Data"][i]["Metadata"][
                        "signal"
                    ]
                )
                or (
                    "FLD" in module["Name"]
                    and "FLD"
                    in intermediate_structured_data["Result Data"][i]["Metadata"][
                        "signal"
                    ]
                )
            ],
            peaks=create_peak(intermediate_structured_data["Result Data"][i]["Peak"])
            if "Peak" in intermediate_structured_data["Result Data"][i]
            else None,
            chromatogram_data_cube=create_data_cube(
                label=f"{intermediate_structured_data['Result Data'][i]['Metadata']['signal'].split(',')[0]} chromatogram",
                measures_value=intermediate_structured_data["Result Data"][i][
                    "Intensity"
                ],
                dimension_value=intermediate_structured_data["Result Data"][i]["Time"],
                data_cube_component=DataCubeComponent(
                    type_=FieldComponentDatatype.double,
                    concept="absorbance"
                    if "DAD"
                    in intermediate_structured_data["Result Data"][i]["Metadata"][
                        "signal"
                    ]
                    else "fluorescence",
                    unit="mAU"
                    if "DAD"
                    in intermediate_structured_data["Result Data"][i]["Metadata"][
                        "signal"
                    ]
                    else "RFU",
                ),
            )
            if "PMP"
            not in intermediate_structured_data["Result Data"][i]["Metadata"]["signal"]
            else None,
        )
        for i in range(intermediate_structured_data["Sample Count"]["count"])
    ]


def create_measurement_groups(
    intermediate_structured_data: dict[str, Any]
) -> MeasurementGroup:
    return MeasurementGroup(
        measurements=create_measurements(intermediate_structured_data),
        measurement_aggregate_custom_info={
            "analytical method identifier": intermediate_structured_data["Result Data"][
                0
            ]["Sample Data"].get("AnalysisMethod")
        },
    )



from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    MeasurementGroup,
)
from allotropy.calcdocs.config import (
    CalcDocsConfig,
    CalculatedDataConfig,
    MeasurementConfig,
)
from allotropy.calcdocs.unchained_labs_lunatic.extractor import LunaticExtractor
from allotropy.calcdocs.unchained_labs_lunatic.views import UuidView
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)


def create_calculated_data(
    measurement_groups: list[MeasurementGroup],
) -> list[CalculatedDocument]:
    elements = []
    for measurement_group in measurement_groups:
        elements += LunaticExtractor.get_elements(measurement_group.measurements)

    lunatic_view_data = UuidView().apply(elements)

    absorbance_conf = MeasurementConfig(
        name="absorbance",
        value="absorbance",
    )

    configs = CalcDocsConfig(
        [
            # a260
            CalculatedDataConfig(
                name="Concentration",
                value="a260 concentration (ng/ul)",
                unit="ng/µL",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="Concentration",
                value="concentration (ng/ul)",
                unit="ng/µL",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="Background (A260)",
                value="background (a260)",
                unit="mAU",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="A260/A230",
                value="a260/a230",
                unit=UNITLESS,
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="A260/A280",
                value="a260/a280",
                unit=UNITLESS,
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            # a280
            CalculatedDataConfig(
                name="Concentration",
                value="concentration (mg/ml)",
                unit="mg/mL",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="Background (A280)",
                value="background (a280)",
                unit="mAU",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="A260/A280",
                value="a260/a280",
                unit=UNITLESS,
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
        ]
    )

    return [
        calc_doc
        for parent_calc_doc in configs.construct()
        for calc_doc in parent_calc_doc.iter_struct()
    ]



from __future__ import annotations

from io import BytesIO
from zipfile import ZipFile

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.strict_xml_element import StrictXmlElement
from allotropy.parsers.utils.zip_handler import ZipHandler


def fix_zip(data: BytesIO) -> BytesIO:
    # ZipFile can fail if there are excess trailing bytes. This function detects the end of zip's central directory and,
    # if found, truncates the data after the end of the zip file.
    content = data.read()
    pos = content.rfind(
        b"\x50\x4b\x05\x06"
    )  # reverse find: this string of bytes is the end of the zip's central directory.
    if pos > 0:
        data.seek(
            pos + 22
        )  # End bytes size is 22, see https://pkwaredownloads.blob.core.windows.net/pkware-general/Documentation/APPNOTE-6.3.0.TXT
        data.truncate()
        data.seek(0)

    return data


class UnicornZipHandler(ZipHandler):
    def get_zip_file(self, data: BytesIO) -> ZipFile:
        return ZipFile(fix_zip(data))

    def get_zip(self, inner_path: str) -> UnicornZipHandler:
        return UnicornZipHandler(self.get_file(inner_path))

    def get_zip_from_pattern(self, pattern: str) -> UnicornZipHandler:
        return self.get_zip(self.get_inner_path(pattern))

    @classmethod
    def create_from_path(cls, path: str) -> UnicornZipHandler:
        with open(path, "rb") as f:
            data = f.read()
        return UnicornZipHandler(data=BytesIO(data))

    def filter_xml_metadata(self, stream: BytesIO) -> BytesIO:
        data = stream.read()
        start = -1
        for idx, element in enumerate(data):
            if int(element) == 60:  # 60 == '<' ASCII char
                start = idx
                break

        end = -1
        for idx, element in enumerate(reversed(data)):
            if int(element) == 62:  # 62 == '>' ASCII char
                end = len(data) - idx
                break

        if start == -1 or end == -1:
            msg = "Unable to extract XML from file."
            raise AllotropeConversionError(msg)

        return BytesIO(data[start:end])

    def get_system_data(self) -> StrictXmlElement:
        system_data = self.get_zip_from_pattern("SystemData(.zip)?$")
        b_stream = system_data.get_file_from_pattern("^Xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_results(self) -> StrictXmlElement:
        b_stream = self.get_file_from_pattern("Result.xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_instrument_config_data(self) -> StrictXmlElement:
        instrument_regex = "InstrumentConfigurationData(.zip)?$"
        instrument_config_data = self.get_zip_from_pattern(instrument_regex)
        b_stream = instrument_config_data.get_file_from_pattern("^Xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_evaluation_log(self) -> StrictXmlElement:
        b_stream = self.get_file_from_pattern("EvaluationLog.xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_chrom_1(self) -> StrictXmlElement:
        b_stream = self.get_file_from_pattern("Chrom.1.Xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_column_type_data(self) -> StrictXmlElement:
        column_type_data = self.get_zip_from_pattern("ColumnTypeData(.zip)?$")
        b_stream = column_type_data.get_file_from_pattern("^Xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)



from __future__ import annotations

from collections import defaultdict

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.view import View


class BlankRoleTypeView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="sample_role_type", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            sample_role_type = element.get_str("sample_role_type")
            if sample_role_type == SampleRoleType.blank_role.name:
                items[str(sample_role_type)].append(element)
        return dict(items)


class CorrectedView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="corrected_value", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            if corrected_value := element.get_float_or_none("corrected_value"):
                items[str(corrected_value)].append(element)
        return dict(items)



from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    ChromatogramDataCube,
    ChromatographyColumnDocument,
    DataProcessingAggregateDocument,
    DataSystemDocument,
    DerivedColumnPressureDataCube,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceDocumentItem,
    DeviceSystemDocument,
    FractionAggregateDocument,
    FractionDocumentItem,
    InjectionDocument,
    LiquidChromatographyAggregateDocument,
    LiquidChromatographyDocumentItem,
    LogAggregateDocument,
    LogDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocument,
    Model,
    Peak as PeakDocument,
    PeakList,
    PostColumnPressureDataCube,
    PreColumnPressureDataCube,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SampleFlowRateDataCube,
    SamplePressureDataCube,
    SolventConcentrationDataCube,
    SystemFlowRateDataCube,
    SystemPressureDataCube,
    TemperatureProfileDataCube,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCentimeter,
    TQuantityValueCubicMillimeter,
    TQuantityValueHertz,
    TQuantityValueMicroliter,
    TQuantityValueMicrometer,
    TQuantityValueMilliAbsorbanceUnitTimesSecond,
    TQuantityValueMilliliter,
    TQuantityValueMilliliterPerMinute,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValuePercent,
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import TDatacube
from allotropy.allotrope.models.shared.definitions.units import SecondTime
from allotropy.allotrope.schema_mappers.data_cube import (
    DataCube,
    get_data_cube,
)
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import quantity_or_none, quantity_or_none_from_unit


@dataclass(frozen=True)
class DeviceDocument:
    device_type: str
    device_identifier: str | None = None
    product_manufacturer: str | None = None
    model_number: str | None = None
    equipment_serial_number: str | None = None
    firmware_version: str | None = None
    device_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Metadata:
    asset_management_identifier: str
    analyst: str | None = None
    detection_type: str | None = None
    model_number: str | None = None
    software_name: str | None = None
    file_name: str | None = None
    data_system_instance_identifier: str | None = None
    unc_path: str | None = None
    equipment_serial_number: str | None = None
    software_version: str | None = None
    product_manufacturer: str | None = None
    brand_name: str | None = None
    device_identifier: str | None = None
    firmware_version: str | None = None
    description: str | None = None
    device_documents: list[DeviceDocument] | None = None
    lc_agg_custom_info: dict[str, Any] | None = None
    data_system_custom_info: dict[str, Any] | None = None
    device_system_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Peak:
    identifier: str
    index: str | None = None
    start: float | None = None
    start_unit: str | None = None
    end: float | None = None
    end_unit: str | None = None
    area: float | None = None
    area_unit: str | None = None
    relative_area: float | None = None
    width: float | None = None
    width_unit: str | None = None
    relative_width: float | None = None
    height: float | None = None
    height_unit: str | None = None
    relative_height: float | None = None
    retention_time: float | None = None
    peak_width_at_5_percent_of_height: float | None = None
    peak_width_at_10_percent_of_height: float | None = None
    peak_width_at_baseline: float | None = None
    asymmetry_factor_measured_at_5_percent_height: float | None = None
    asymmetry_factor_measured_at_10_percent_height: float | None = None
    number_of_theoretical_plates__chromatography_: float | None = None
    written_name: str | None = None
    chromatographic_resolution: float | None = None
    chromatographic_asymmetry: float | None = None
    width_at_half_height: float | None = None
    width_at_half_height_unit: str | None = None
    peak_analyte_amount: float | None = None
    relative_peak_analyte_amount: float | None = None
    custom_info: dict[str, Any] | None = None
    relative_retention_time: float | None = None
    capacity_factor: float | None = None
    number_of_theoretical_plates_by_peak_width_at_half_height: float | None = None

    relative_corrected_peak_area: float | None = None
    peak_group: float | None = None
    baseline_value_at_start_of_peak: float | None = None
    baseline_value_at_end_of_peak: float | None = None


@dataclass(frozen=True)
class Fraction:
    index: str
    fraction_role: str | None = None
    field_type: str | None = None
    retention_time: float | None = None
    retention_volume: float | None = None


@dataclass(frozen=True)
class Log:
    index: str
    method_identifier: str | None = None
    log_entry: str | None = None
    retention_time: float | None = None
    retention_volume: float | None = None


@dataclass(frozen=True)
class DeviceControlDoc:
    device_type: str
    start_time: str | None = None
    detection_type: str | None = None
    device_identifier: str | None = None
    product_manufacturer: str | None = None
    model_number: str | None = None
    equipment_serial_number: str | None = None
    excitation_wavelength_setting: float | None = None
    detector_wavelength_setting: float | None = None
    detector_bandwidth_setting: float | None = None
    device_control_custom_info: dict[str, Any] | None = None
    firmware_version: str | None = None
    solvent_conc_data_cube: DataCube | None = None
    pre_column_pressure_data_cube: DataCube | None = None
    sample_pressure_data_cube: DataCube | None = None
    system_pressure_data_cube: DataCube | None = None
    post_column_pressure_data_cube: DataCube | None = None
    sample_flow_data_cube: DataCube | None = None
    system_flow_data_cube: DataCube | None = None
    temperature_profile_data_cube: DataCube | None = None
    detector_offset_setting: float | None = None
    detector_sampling_rate_setting: float | None = None
    electronic_absorbance_reference_bandwidth_setting: float | None = None
    electronic_absorbance_reference_wavelength_setting: float | None = None


@dataclass(frozen=True)
class ProcessedData:
    custom_info: dict[str, Any] | None = None
    data: list[dict[str, Any]] | None = None


@dataclass(frozen=True)
class Measurement:
    measurement_identifier: str
    sample_identifier: str

    # Injection metadata
    injection_identifier: str
    injection_time: str

    device_control_docs: list[DeviceControlDoc]

    # Optional metadata
    processed_data_identifier: str | None = None
    measurement_time: str | None = None
    sample_role_type: str | None = None
    written_name: str | None = None
    chromatography_serial_num: str | None = None
    chromatography_part_num: str | None = None
    column_inner_diameter: float | None = None
    chromatography_chemistry_type: str | None = None
    chromatography_particle_size: float | None = None
    column_product_manufacturer: str | None = None
    chromatography_length: float | None = None
    batch_identifier: str | None = None
    void_volume: float | None = None
    flow_rate: float | None = None
    description: str | None = None
    location_identifier: str | None = None
    well_location_identifier: str | None = None
    observation: str | None = None
    injection_volume_setting: float | None = None
    autosampler_injection_volume_setting: float | None = None

    # Measurement data cubes
    chromatogram_data_cube: DataCube | None = None
    processed_data_chromatogram_data_cube: DataCube | None = None
    derived_column_pressure_data_cube: DataCube | None = None

    peaks: list[Peak] | None = None
    processed_data: list[ProcessedData] | None = None

    sample_custom_info: dict[str, Any] | None = None
    injection_custom_info: dict[str, Any] | None = None
    column_custom_info: dict[str, Any] | None = None
    measurement_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    fractions: list[Fraction] | None = None
    logs: list[Log] | None = None
    measurement_aggregate_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/liquid-chromatography/BENCHLING/2023/09/liquid-chromatography.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            liquid_chromatography_aggregate_document=add_custom_information_document(
                LiquidChromatographyAggregateDocument(
                    liquid_chromatography_document=[
                        self._get_technique_document(group, data.metadata)
                        for group in data.measurement_groups
                    ],
                    device_system_document=add_custom_information_document(
                        DeviceSystemDocument(
                            asset_management_identifier=data.metadata.asset_management_identifier,
                            product_manufacturer=data.metadata.product_manufacturer,
                            device_identifier=data.metadata.device_identifier,
                            firmware_version=data.metadata.firmware_version,
                            brand_name=data.metadata.brand_name,
                            model_number=data.metadata.model_number,
                            device_document=(
                                [
                                    add_custom_information_document(
                                        DeviceDocumentItem(
                                            device_type=doc.device_type,
                                            device_identifier=doc.device_identifier,
                                            product_manufacturer=doc.product_manufacturer,
                                            model_number=doc.model_number,
                                            equipment_serial_number=doc.equipment_serial_number,
                                            firmware_version=doc.firmware_version,
                                        ),
                                        custom_info_doc=doc.device_custom_info,
                                    )
                                    for doc in data.metadata.device_documents
                                ]
                                if data.metadata.device_documents
                                else None
                            ),
                        ),
                        data.metadata.device_system_custom_info,
                    ),
                    data_system_document=add_custom_information_document(
                        DataSystemDocument(
                            file_name=data.metadata.file_name,
                            data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                            UNC_path=data.metadata.unc_path,
                            software_name=data.metadata.software_name,
                            software_version=data.metadata.software_version,
                            ASM_converter_name=self.converter_name,
                            ASM_converter_version=ASM_CONVERTER_VERSION,
                        ),
                        data.metadata.data_system_custom_info,
                    ),
                ),
                data.metadata.lc_agg_custom_info,
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, group: MeasurementGroup, metadata: Metadata
    ) -> LiquidChromatographyDocumentItem:
        return LiquidChromatographyDocumentItem(
            analyst=metadata.analyst,
            measurement_aggregate_document=add_custom_information_document(
                MeasurementAggregateDocument(
                    measurement_document=[
                        self._get_measurement_document_item(measurement)
                        for measurement in group.measurements
                    ],
                    fraction_aggregate_document=self._get_fraction_aggregate_document(
                        group.fractions
                    ),
                    log_aggregate_document=self._get_log_aggregate_document(group.logs),
                ),
                custom_info_doc=group.measurement_aggregate_custom_info,
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurement
    ) -> MeasurementDocument:
        if len(measurement.device_control_docs) < 1:
            msg = "Expected at least one device control document in measurement."
            raise AllotropeConversionError(msg)

        return add_custom_information_document(
            MeasurementDocument(
                measurement_identifier=measurement.measurement_identifier,
                measurement_time=self.get_date_time(measurement.measurement_time)
                if measurement.measurement_time is not None
                else None,
                chromatography_column_document=self._get_chromatography_column_document(
                    measurement
                ),
                sample_document=self._get_sample_document(measurement),
                injection_document=(
                    self._get_injection_document(measurement)
                    if measurement.injection_identifier
                    else None
                ),
                processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                    measurement
                ),
                device_control_aggregate_document=DeviceControlAggregateDocument(
                    device_control_document=[
                        self._get_device_control_document(device_control_doc)
                        for device_control_doc in measurement.device_control_docs
                    ]
                ),
                chromatogram_data_cube=(
                    get_data_cube(
                        measurement.chromatogram_data_cube,
                        ChromatogramDataCube,
                    )
                ),
            ),
            measurement.measurement_custom_info,
        )

    def _get_chromatography_column_document(
        self, measurement: Measurement
    ) -> ChromatographyColumnDocument:
        return add_custom_information_document(
            ChromatographyColumnDocument(
                chromatography_column_serial_number=measurement.chromatography_serial_num,
                chromatography_column_chemistry_type=measurement.chromatography_chemistry_type,
                column_inner_diameter=quantity_or_none(
                    TQuantityValueMillimeter,
                    measurement.column_inner_diameter,
                ),
                chromatography_column_part_number=measurement.chromatography_part_num,
                chromatography_column_particle_size=quantity_or_none(
                    TQuantityValueMicrometer, measurement.chromatography_particle_size
                ),
                chromatography_column_length=quantity_or_none(
                    TQuantityValueCentimeter, measurement.chromatography_length
                ),
                product_manufacturer=measurement.column_product_manufacturer,
                void_volume=quantity_or_none(
                    TQuantityValueMilliliter, measurement.void_volume
                ),
            ),
            custom_info_doc=measurement.column_custom_info,
        )

    def _get_injection_document(self, measurement: Measurement) -> InjectionDocument:
        return add_custom_information_document(
            InjectionDocument(
                injection_identifier=measurement.injection_identifier,
                injection_time=self.get_date_time(measurement.injection_time),
                autosampler_injection_volume_setting__chromatography_=quantity_or_none(
                    TQuantityValueCubicMillimeter,
                    measurement.autosampler_injection_volume_setting,
                ),
                injection_volume_setting=quantity_or_none(
                    TQuantityValueMicroliter, measurement.injection_volume_setting
                ),
            ),
            measurement.injection_custom_info,
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:

        return add_custom_information_document(
            SampleDocument(
                sample_identifier=measurement.sample_identifier,
                batch_identifier=measurement.batch_identifier,
                description=measurement.description,
                sample_role_type=measurement.sample_role_type,
                written_name=measurement.written_name,
                flow_rate=quantity_or_none(
                    TQuantityValueMilliliterPerMinute, measurement.flow_rate
                ),
                location_identifier=measurement.location_identifier,
                well_location_identifier=measurement.well_location_identifier,
                observation=measurement.observation,
            ),
            measurement.sample_custom_info,
        )

    def _get_peak_document(self, peak: Peak) -> PeakDocument:
        return add_custom_information_document(
            PeakDocument(
                identifier=peak.identifier,
                peak_index=peak.index,
                peak_start=quantity_or_none_from_unit(peak.start_unit, peak.start),  # type: ignore[arg-type]
                peak_end=quantity_or_none_from_unit(peak.end_unit, peak.end),  # type: ignore[arg-type]
                peak_area=quantity_or_none_from_unit(peak.area_unit, peak.area),
                peak_width=quantity_or_none_from_unit(peak.width_unit, peak.width),  # type: ignore[arg-type]
                peak_height=quantity_or_none_from_unit(peak.height_unit, peak.height),
                relative_peak_area=quantity_or_none(
                    TQuantityValuePercent, peak.relative_area
                ),
                relative_peak_height=quantity_or_none(
                    TQuantityValuePercent, peak.relative_height
                ),
                retention_time=quantity_or_none(
                    TQuantityValueSecondTime, peak.retention_time
                ),
                chromatographic_peak_resolution=quantity_or_none(
                    TQuantityValueUnitless, peak.chromatographic_resolution
                ),
                chromatographic_peak_asymmetry_factor=quantity_or_none(
                    TQuantityValueUnitless, peak.chromatographic_asymmetry
                ),
                peak_width_at_half_height=quantity_or_none_from_unit(  # type: ignore[arg-type]
                    peak.width_at_half_height_unit or SecondTime.unit,
                    peak.width_at_half_height,
                ),
                relative_retention_time=quantity_or_none(
                    TQuantityValuePercent, peak.relative_retention_time
                ),
                capacity_factor__chromatography_=quantity_or_none(
                    TQuantityValueUnitless, peak.capacity_factor
                ),
                number_of_theoretical_plates_by_peak_width_at_half_height=quantity_or_none(
                    TQuantityValueUnitless,
                    peak.number_of_theoretical_plates_by_peak_width_at_half_height,
                ),
                peak_width_at_5___of_height=quantity_or_none(
                    TQuantityValueSecondTime, peak.peak_width_at_5_percent_of_height
                ),
                peak_width_at_10___of_height=quantity_or_none(
                    TQuantityValueSecondTime, peak.peak_width_at_10_percent_of_height
                ),
                peak_width_at_baseline=quantity_or_none(
                    TQuantityValueSecondTime, peak.peak_width_at_baseline
                ),
                asymmetry_factor_measured_at_5___height=quantity_or_none(
                    TQuantityValueUnitless,
                    peak.asymmetry_factor_measured_at_5_percent_height,
                ),
                asymmetry_factor_measured_at_10___height=quantity_or_none(
                    TQuantityValueUnitless,
                    peak.asymmetry_factor_measured_at_10_percent_height,
                ),
                number_of_theoretical_plates__chromatography_=quantity_or_none(
                    TQuantityValueUnitless,
                    peak.number_of_theoretical_plates__chromatography_,
                ),
                written_name=peak.written_name,
                peak_analyte_amount=quantity_or_none(
                    TQuantityValueUnitless, peak.peak_analyte_amount
                ),
                relative_corrected_peak_area=quantity_or_none(
                    TQuantityValuePercent, peak.relative_corrected_peak_area
                ),
                peak_group=quantity_or_none(
                    TQuantityValueMilliAbsorbanceUnitTimesSecond, peak.peak_group
                ),
                baseline_value_at_start_of_peak=quantity_or_none(
                    TQuantityValueSecondTime, peak.baseline_value_at_start_of_peak
                ),
                baseline_value_at_end_of_peak=quantity_or_none(
                    TQuantityValueSecondTime, peak.baseline_value_at_end_of_peak
                ),
                relative_peak_analyte_amount=quantity_or_none(
                    TQuantityValuePercent, peak.relative_peak_analyte_amount
                ),
            ),
            peak.custom_info,
        )

    def _get_processed_data_aggregate_document(
        self, measurement: Measurement
    ) -> ProcessedDataAggregateDocument | None:
        if not (
            measurement.processed_data_chromatogram_data_cube
            or measurement.derived_column_pressure_data_cube
            or measurement.peaks
        ):
            return None

        def get_data_processing_documents(pdd: ProcessedData) -> list[dict[str, Any]]:
            if not pdd.data:
                return []
            return pdd.data

        def build_peak_list() -> PeakList | None:
            if not measurement.peaks:
                return None
            return PeakList(
                peak=[self._get_peak_document(p) for p in measurement.peaks]
            )

        def build_data_document_item(
            pdd: ProcessedData | None = None,
        ) -> ProcessedDataDocumentItem:
            data_processing = (
                DataProcessingAggregateDocument(
                    data_processing_document=get_data_processing_documents(pdd)
                )
                if pdd and get_data_processing_documents(pdd)
                else None
            )
            item = ProcessedDataDocumentItem(
                chromatogram_data_cube=get_data_cube(
                    measurement.processed_data_chromatogram_data_cube, TDatacube
                ),
                derived_column_pressure_data_cube=get_data_cube(
                    measurement.derived_column_pressure_data_cube,
                    DerivedColumnPressureDataCube,
                ),
                processed_data_identifier=measurement.processed_data_identifier,
                peak_list=build_peak_list(),
                data_processing_aggregate_document=data_processing,
            )
            return (
                add_custom_information_document(item, pdd.custom_info) if pdd else item
            )

        if measurement.processed_data:
            processed_data_documents = [
                build_data_document_item(pdd) for pdd in measurement.processed_data
            ]
        else:
            processed_data_documents = [build_data_document_item()]

        return ProcessedDataAggregateDocument(
            processed_data_document=processed_data_documents
        )

    def _get_device_control_document(
        self, device_control_doc: DeviceControlDoc
    ) -> DeviceControlDocumentItem:
        custom_info = device_control_doc.device_control_custom_info or {}

        return add_custom_information_document(
            DeviceControlDocumentItem(
                device_type=device_control_doc.device_type,
                start_time_setting=(
                    self.get_date_time(device_control_doc.start_time)
                    if device_control_doc.start_time is not None
                    else None
                ),
                device_identifier=device_control_doc.device_identifier,
                detection_type=device_control_doc.detection_type,
                product_manufacturer=device_control_doc.product_manufacturer,
                equipment_serial_number=device_control_doc.equipment_serial_number,
                model_number=device_control_doc.model_number,
                firmware_version=device_control_doc.firmware_version,
                excitation_wavelength_setting=quantity_or_none(
                    TQuantityValueNanometer,
                    device_control_doc.excitation_wavelength_setting,
                ),
                detector_wavelength_setting=quantity_or_none(
                    TQuantityValueNanometer,
                    device_control_doc.detector_wavelength_setting,
                ),
                detector_bandwidth_setting=quantity_or_none(
                    TQuantityValueNanometer,
                    device_control_doc.detector_bandwidth_setting,
                ),
                solvent_concentration_data_cube=get_data_cube(
                    device_control_doc.solvent_conc_data_cube,
                    SolventConcentrationDataCube,
                ),
                pre_column_pressure_data_cube=get_data_cube(
                    device_control_doc.pre_column_pressure_data_cube,
                    PreColumnPressureDataCube,
                ),
                sample_pressure_data_cube=get_data_cube(
                    device_control_doc.sample_pressure_data_cube,
                    SamplePressureDataCube,
                ),
                system_pressure_data_cube=get_data_cube(
                    device_control_doc.system_pressure_data_cube,
                    SystemPressureDataCube,
                ),
                post_column_pressure_data_cube=get_data_cube(
                    device_control_doc.post_column_pressure_data_cube,
                    PostColumnPressureDataCube,
                ),
                sample_flow_rate_data_cube=get_data_cube(
                    device_control_doc.sample_flow_data_cube,
                    SampleFlowRateDataCube,
                ),
                system_flow_rate_data_cube=get_data_cube(
                    device_control_doc.system_flow_data_cube,
                    SystemFlowRateDataCube,
                ),
                temperature_profile_data_cube=get_data_cube(
                    device_control_doc.temperature_profile_data_cube,
                    TemperatureProfileDataCube,
                ),
                detector_offset_setting=quantity_or_none(
                    TQuantityValueUnitless, device_control_doc.detector_offset_setting
                ),
                detector_sampling_rate_setting=quantity_or_none(
                    TQuantityValueHertz,
                    device_control_doc.detector_sampling_rate_setting,
                ),
                electronic_absorbance_reference_bandwidth_setting=quantity_or_none(
                    TQuantityValueNanometer,
                    device_control_doc.electronic_absorbance_reference_bandwidth_setting,
                ),
                electronic_absorbance_reference_wavelength_setting=quantity_or_none(
                    TQuantityValueNanometer,
                    device_control_doc.electronic_absorbance_reference_wavelength_setting,
                ),
            ),
            custom_info_doc=custom_info,
        )

    def _get_fraction_aggregate_document(
        self, fractions: list[Fraction] | None
    ) -> FractionAggregateDocument | None:
        if fractions is None:
            return None

        return FractionAggregateDocument(
            fraction_document=[
                self._get_fraction_document(fraction_doc)
                for fraction_doc in fractions or []
            ]
        )

    def _get_fraction_document(self, fraction_doc: Fraction) -> FractionDocumentItem:
        return FractionDocumentItem(
            index=fraction_doc.index,
            fraction_role=fraction_doc.fraction_role,
            field_type=fraction_doc.field_type,
            retention_time=quantity_or_none(
                TQuantityValueSecondTime, fraction_doc.retention_time
            ),
            retention_volume=quantity_or_none(
                TQuantityValueMilliliter, fraction_doc.retention_volume
            ),
        )

    def _get_log_aggregate_document(
        self, logs: list[Log] | None
    ) -> LogAggregateDocument | None:
        if logs is None:
            return None

        return LogAggregateDocument(
            log_document=[self._get_log_document(log_doc) for log_doc in logs or []]
        )

    def _get_log_document(self, log_doc: Log) -> LogDocumentItem:
        return LogDocumentItem(
            index=log_doc.index,
            method_identifier=log_doc.method_identifier,
            log_entry=log_doc.log_entry,
            retention_time=quantity_or_none(
                TQuantityValueSecondTime, log_doc.retention_time
            ),
            retention_volume=quantity_or_none(
                TQuantityValueMilliliter, log_doc.retention_volume
            ),
        )



NUCLEOCOUNTER_SOFTWARE_NAME = "NucleoView"
NUCLEOCOUNTER_DEVICE_TYPE = "dark field imager (cell counter)"
NUCLEOCOUNTER_DETECTION_TYPE = "dark field"

DEFAULT_ANALYST = "NC-200"
DEFAULT_MODEL_NUMBER = "NucleoCounter NC-200"

MEASUREMENT_AGG_DOCUMENT_CUSTOM_FIELDS = {
    "Protocol template purpose",
    "Protocol template filename",
    "Protocol template title",
    "Protocol adaptation purpose",
    "Protocol adaptation title",
    "Protocol adaptation filename",
    "Login ID",
    "Event log format",
}



from __future__ import annotations

from pathlib import Path
from typing import Any

import pandas as pd

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDalton,
    TQuantityValueHertz,
    TQuantityValueMicroliterPerMinute,
    TQuantityValueMilliliter,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import ResonanceUnits, Unitless
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    DeviceDocument,
    Measurement,
    MeasurementGroup,
    Metadata,
    ReportPoint,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.calcdocs.config import (
    CalcDocsConfig,
    CalculatedDataConfig,
    MeasurementConfig,
)
from allotropy.calcdocs.cytiva_biacore_t200_control.extractor import (
    CytivaBiacoreExtractor,
)
from allotropy.calcdocs.cytiva_biacore_t200_control.views import ReportPointDataView
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.cytiva_biacore_t200_control import constants
from allotropy.parsers.cytiva_biacore_t200_control.cytiva_biacore_t200_control_structure import (
    Data,
    MeasurementData,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)
from allotropy.parsers.utils.values import quantity_or_none, try_float_or_none
from allotropy.types import DictType


def _get_sensorgram_datacube(sensorgram_data: pd.DataFrame) -> DataCube:
    cycle = sensorgram_data.iloc[0]["Cycle Number"]
    flow_cell = sensorgram_data.iloc[0]["Flow Cell Number"]
    return DataCube(
        label=f"Cycle{cycle}_FlowCell{flow_cell}",
        structure_dimensions=[
            DataCubeComponent(FieldComponentDatatype.double, "elapsed time", "s")
        ],
        structure_measures=[
            DataCubeComponent(FieldComponentDatatype.double, "resonance", "RU")
        ],
        dimensions=[sensorgram_data["Time (s)"].astype(float).to_list()],
        measures=[sensorgram_data["Sensorgram (RU)"].astype(float).to_list()],
    )


def _get_device_control_custom_info(data: Data) -> DictType:
    custom_ifo: dict[str, Any] = {
        "number of flow cells": data.chip_data.number_of_flow_cells,
        "number of spots": data.chip_data.number_of_spots,
        "buffer volume": quantity_or_none(
            TQuantityValueMilliliter, data.run_metadata.buffer_volume
        ),
    }
    if detection_setting := data.run_metadata.detection_setting:
        custom_ifo.update({detection_setting.key: detection_setting.value})
    return custom_ifo


def create_metadata(data: Data, named_file_contents: NamedFileContents) -> Metadata:
    filepath = Path(named_file_contents.original_file_path)
    run_metadata = data.run_metadata
    system_information = data.system_information
    chip_data = data.chip_data

    return Metadata(
        brand_name=constants.BRAND_NAME,
        device_identifier=system_information.device_identifier,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        asm_file_identifier=filepath.with_suffix(".json").name,
        model_number=system_information.model_number,
        data_system_instance_identifier=NOT_APPLICABLE,
        file_name=filepath.name,
        unc_path=named_file_contents.original_file_path,
        software_name=system_information.software_name,
        software_version=system_information.software_version,
        detection_type=constants.SURFACE_PLASMON_RESONANCE,
        compartment_temperature=run_metadata.compartment_temperature,
        sensor_chip_type=chip_data.sensor_chip_type,
        lot_number=chip_data.lot_number,
        sensor_chip_identifier=chip_data.sensor_chip_identifier,
        device_document=[
            DeviceDocument(device.type_, device.identifier)
            for device in run_metadata.devices
        ],
        sensor_chip_custom_info=chip_data.custom_info,
    )


def create_measurements(
    measurements_data: list[MeasurementData], device_control_custom_info: DictType
) -> list[Measurement]:
    return [
        Measurement(
            identifier=measurement.identifier,
            type_=measurement.type_,
            device_type=measurement.device_type,
            sample_identifier=measurement.sample_identifier,
            location_identifier=measurement.location_identifier,
            sample_role_type=measurement.sample_role_type,
            concentration=measurement.concentration,
            flow_cell_identifier=measurement.flow_cell_identifier,
            device_control_custom_info=device_control_custom_info,
            sample_custom_info={
                "molecular weight": quantity_or_none(
                    TQuantityValueDalton, measurement.molecular_weight
                )
            },
            sensorgram_data_cube=_get_sensorgram_datacube(measurement.sensorgram_data),
            report_point_data=(
                [
                    ReportPoint(
                        identifier=rp_data.identifier,
                        identifier_role=rp_data.identifier_role,
                        absolute_resonance=rp_data.absolute_resonance,
                        relative_resonance=rp_data.relative_resonance,
                        time_setting=rp_data.time_setting,
                        custom_info=rp_data.custom_info,
                    )
                    for rp_data in measurement.report_point_data
                ]
                if measurement.report_point_data is not None
                else None
            ),
            # for Mobilization experiments
            method_name=measurement.method_name,
            ligand_identifier=measurement.ligand_identifier,
            flow_path=measurement.flow_path,
            flow_rate=measurement.flow_rate,
            contact_time=measurement.contact_time,
            dilution=measurement.dilution,
        )
        for measurement in measurements_data
    ]


def create_measurement_groups(data: Data) -> list[MeasurementGroup]:
    system_information = data.system_information
    device_control_custom_info = _get_device_control_custom_info(data)
    return [
        MeasurementGroup(
            measurement_time=system_information.measurement_time,
            measurements=create_measurements(
                measurements_data, device_control_custom_info
            ),
            experiment_type=system_information.experiment_type,
            analytical_method_identifier=system_information.analytical_method_identifier,
            analyst=data.run_metadata.analyst,
            measurement_aggregate_custom_info={
                "baseline flow": quantity_or_none(
                    TQuantityValueMicroliterPerMinute, data.run_metadata.baseline_flow
                ),
                "data collection rate": quantity_or_none(
                    TQuantityValueHertz,
                    try_float_or_none(data.run_metadata.data_collection_rate),
                ),
            },
        )
        for measurements_data in data.sample_data.measurements.values()
    ]


def create_calculated_data(data: Data) -> list[CalculatedDocument]:
    report_point_data_view = ReportPointDataView().apply(
        CytivaBiacoreExtractor.sample_data_to_elements(data.sample_data)
    )
    absolute_resonance_conf = MeasurementConfig(
        name="Absolute Resonance",
        value="absolute_resonance",
    )

    configs = CalcDocsConfig(
        [
            CalculatedDataConfig(
                name="Min Resonance",
                value="min_resonance",
                view_data=report_point_data_view,
                source_configs=(absolute_resonance_conf,),
                unit=ResonanceUnits.unit,
            ),
            CalculatedDataConfig(
                name="Max Resonance",
                value="max_resonance",
                view_data=report_point_data_view,
                source_configs=(absolute_resonance_conf,),
                unit=ResonanceUnits.unit,
            ),
            CalculatedDataConfig(
                name="LRSD",
                value="lrsd",
                view_data=report_point_data_view,
                source_configs=(absolute_resonance_conf,),
                unit=Unitless.unit,
            ),
            CalculatedDataConfig(
                name="Slope",
                value="slope",
                view_data=report_point_data_view,
                source_configs=(absolute_resonance_conf,),
                unit=Unitless.unit,
            ),
            CalculatedDataConfig(
                name="SD",
                value="sd",
                view_data=report_point_data_view,
                source_configs=(absolute_resonance_conf,),
                unit=Unitless.unit,
            ),
        ]
    )

    return [
        calc_doc
        for parent_calc_doc in configs.construct()
        for calc_doc in parent_calc_doc.iter_struct()
    ]



from __future__ import annotations

from enum import Enum
import math
from pathlib import Path

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueSecondTime,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)
from allotropy.allotrope.schema_mappers.adm.flow_cytometry.benchling._2025._03.flow_cytometry import (
    CompensationMatrix,
    CompensationMatrixGroup,
    DataRegion,
    Measurement,
    MeasurementGroup,
    Metadata,
    Population,
    Statistic,
    StatisticDimension,
    Vertex,
)
from allotropy.exceptions import AllotropeParsingError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.flowjo import constants
from allotropy.parsers.utils.strict_xml_element import StrictXmlElement
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float_or_none

# Map of FlowJo field names to statistic datum roles and units
FLOWJO_STATISTIC_MAP = {
    # Fluorescence statistics
    "Median": {"role": TStatisticDatumRole.median_role.value, "unit": "RFU"},
    "CV": {
        "role": TStatisticDatumRole.coefficient_of_variation_role.value,
        "unit": "%",
    },
    "Robust CV": {
        "role": TStatisticDatumRole.robust_coefficient_of_variation_role.value,
        "unit": "%",
    },
    "Mean": {"role": TStatisticDatumRole.arithmetic_mean_role.value, "unit": "RFU"},
    "Geometric Mean": {
        "role": TStatisticDatumRole.geometric_mean_role.value,
        "unit": "RFU",
    },
    "Percentile": {"role": TStatisticDatumRole.percentile_role.value, "unit": "RFU"},
    "SD": {
        "role": TStatisticDatumRole.standard_deviation_role.value,
        "unit": "(unitless)",
    },
    "MADExact": {
        "role": TStatisticDatumRole.median_absolute_deviation_percentile_role.value,
        "unit": "(unitless)",
    },
    "Robust SD": {
        "role": TStatisticDatumRole.robust_standard_deviation_role.value,
        "unit": "(unitless)",
    },
    "Median Abs Dev": {
        "role": TStatisticDatumRole.median_absolute_deviation_role.value,
        "unit": "(unitless)",
    },
    # Count statistics
    "fj.stat.freqofparent": {
        "role": TStatisticDatumRole.frequency_of_parent_role.value,
        "unit": "%",
    },
    "fj.stat.freqofgrandparent": {
        "role": TStatisticDatumRole.frequency_of_grandparent_role.value,
        "unit": "%",
    },
    "fj.stat.freqoftotal": {
        "role": TStatisticDatumRole.frequency_of_total_role.value,
        "unit": "%",
    },
}

# Identify statistics that belong to the "Count" feature
COUNT_FIELDS = [
    "fj.stat.freqofparent",
    "fj.stat.freqofgrandparent",
    "fj.stat.freqoftotal",
    "fj.stat.freqof",
]
FLUORESCENCE_FIELDS = [
    "Median",
    "CV",
    "Robust CV",
    "Mean",
    "Geometric Mean",
    "Percentile",
    "SD",
    "MADExact",
    "Robust SD",
    "Median Abs Dev",
]


class RegionType(Enum):
    RECTANGLE = "Rectangle"
    POLYGON = "Polygon"
    ELLIPSOID = "Ellipsoid"
    CURLY_QUAD = "CurlyQuad"


class VertexRole(Enum):
    FOCI = "foci"
    EDGE = "edge"


class VertexExtractor:
    """Base class for vertex extraction strategies."""

    def __init__(
        self, gate_element: StrictXmlElement, x_dim: str | None, y_dim: str | None
    ):
        self.gate_element = gate_element
        self.x_dim = x_dim
        self.y_dim = y_dim
        self.x_unit = (
            TQuantityValueSecondTime.unit
            if x_dim is not None and x_dim.lower() == "time"
            else TQuantityValueRelativeFluorescenceUnit.unit
        )
        self.y_unit = (
            TQuantityValueSecondTime.unit
            if y_dim is not None and y_dim.lower() == "time"
            else TQuantityValueRelativeFluorescenceUnit.unit
        )

    def extract(self) -> list[Vertex] | None:
        """Extract vertices from the gate element."""
        raise NotImplementedError

    @staticmethod
    def build(
        gate_element: StrictXmlElement,
        gate_type: str | None,
        x_dim: str | None,
        y_dim: str | None,
    ) -> VertexExtractor | None:
        if not gate_type:
            return None

        extractors = {
            RegionType.POLYGON.value: PolygonVertexExtractor,
            RegionType.RECTANGLE.value: RectangleVertexExtractor,
            RegionType.CURLY_QUAD.value: RectangleVertexExtractor,
            RegionType.ELLIPSOID.value: EllipsoidVertexExtractor,
        }

        extractor = extractors.get(gate_type)
        if not extractor:
            msg = f"Gate type '{gate_type}' is not currently supported"
            raise AllotropeParsingError(msg)

        strategy = extractor(gate_element, x_dim, y_dim)
        return strategy


class PolygonVertexExtractor(VertexExtractor):
    """Strategy for extracting vertices from Polygon gates."""

    def extract(self) -> list[Vertex] | None:
        result = []
        vertex_elements = self.gate_element.findall("gating:vertex")

        if not vertex_elements:
            return None

        for vertex in vertex_elements:
            coordinates = vertex.findall("gating:coordinate")
            if len(coordinates) < 2:
                return None

            x = coordinates[0].get_namespaced_attr_or_none("data-type", "value")
            y = coordinates[1].get_namespaced_attr_or_none("data-type", "value")

            if x is not None and y is not None:
                result.append(
                    Vertex(
                        x_coordinate=float(x),
                        y_coordinate=float(y),
                        x_unit=self.x_unit,
                        y_unit=self.y_unit,
                    )
                )

        return result if result else None


class RectangleVertexExtractor(VertexExtractor):
    """Strategy for extracting vertices from Rectangle and CurlyQuad gates."""

    def extract(self) -> list[Vertex] | None:
        result = []
        dimension_elements = self.gate_element.findall("gating:dimension")

        if len(dimension_elements) < 2:
            return None

        def _get_gate_value_from_dimension(
            dimension: StrictXmlElement, gate_type: str
        ) -> str | None:
            element = dimension.find_or_none(f"gating:{gate_type}")
            return (
                element.get_namespaced_attr_or_none("data-type", "value")
                if element
                else dimension.get_namespaced_attr_or_none("gating", gate_type)
            )

        x_min = _get_gate_value_from_dimension(dimension_elements[0], "min")
        x_max = _get_gate_value_from_dimension(dimension_elements[0], "max")
        y_min = _get_gate_value_from_dimension(dimension_elements[1], "min")
        y_max = _get_gate_value_from_dimension(dimension_elements[1], "max")

        for x, y in ((x_min, y_min), (x_min, y_max), (x_max, y_max), (x_max, y_min)):
            if x is not None and y is not None:
                result.append(
                    Vertex(
                        x_coordinate=float(x),
                        y_coordinate=float(y),
                        x_unit=self.x_unit,
                        y_unit=self.y_unit,
                    )
                )

        return result if result else None


class EllipsoidVertexExtractor(VertexExtractor):
    """Strategy for extracting vertices from Ellipsoid gates."""

    def extract(self) -> list[Vertex] | None:
        result = []

        def _extract_coordinate_values(
            coords: list[StrictXmlElement],
        ) -> tuple[str | None, str | None]:
            x_value = (
                coords[0].get_namespaced_attr_or_none("data-type", "value")
                if coords
                else None
            )
            y_value = (
                coords[1].get_namespaced_attr_or_none("data-type", "value")
                if len(coords) > 1
                else None
            )
            return x_value, y_value

        def _add_vertex_with_role(vertex_element: StrictXmlElement, role: str) -> None:
            coords = vertex_element.findall("gating:coordinate")
            x_value, y_value = _extract_coordinate_values(coords)

            if x_value is not None and y_value is not None:
                result.append(
                    Vertex(
                        x_coordinate=float(x_value),
                        y_coordinate=float(y_value),
                        x_unit=self.x_unit,
                        y_unit=self.y_unit,
                        vertex_role=role,
                    )
                )

        foci_vertices = self.gate_element.findall("gating:foci/gating:vertex")
        for vertex in foci_vertices:
            _add_vertex_with_role(vertex, VertexRole.FOCI.value)

        edge_vertices = self.gate_element.findall("gating:edge/gating:vertex")
        for vertex in edge_vertices:
            _add_vertex_with_role(vertex, VertexRole.EDGE.value)

        return result if result else None


def create_metadata(root_element: StrictXmlElement, file_path: str) -> Metadata:
    cytometer = root_element.recursive_find_or_none(["Cytometers", "Cytometer"])
    keywords = root_element.recursive_find_or_none(["SampleList", "Sample", "Keywords"])

    equipment_serial_number = None
    if keywords is not None:
        keyword_elements = keywords.findall("Keyword")
        for kw in keyword_elements:
            name = kw.get_attr_or_none("name")
            if name in ["CTNUM", "$CYTSN"]:
                equipment_serial_number = kw.get_attr_or_none("value")
                break

    return Metadata(
        file_name=_get_file_path(root_element.get_attr_or_none("nonAutoSaveFileName"))
        or Path(file_path).name,
        unc_path=file_path,
        device_identifier=constants.DEVICE_IDENTIFIER,
        model_number=cytometer.get_attr_or_none("cyt") if cytometer else None,
        equipment_serial_number=equipment_serial_number,
        data_system_instance_identifier=NOT_APPLICABLE,
        software_name=constants.SOFTWARE_NAME,
        software_version=root_element.get_attr_or_none("flowJoVersion"),
        asm_file_identifier=Path(file_path).with_suffix(".json").name,
    )


def _get_file_path(file_path: str | None) -> str | None:
    if file_path is None:
        return None
    return file_path.split("/")[-1]


def _get_keyword_value_by_name_from_sample(
    sample: StrictXmlElement, name: str
) -> str | None:
    keywords = sample.find_or_none("Keywords")
    if keywords is None:
        return None

    keyword_elements = keywords.findall("Keyword")
    for kw in keyword_elements:
        if kw.get_attr_or_none("name") == name:
            return kw.get_attr_or_none("value")
    return None


def _create_compensation_matrix_groups(
    sample: StrictXmlElement,
) -> list[CompensationMatrixGroup] | None:
    transform_matrix_element = sample.find_or_none("transforms:spilloverMatrix")
    if transform_matrix_element is None:
        return None

    transform_spillovers = transform_matrix_element.findall("transforms:spillover")
    if not transform_spillovers:
        return None

    result = []
    for transform_spillover in transform_spillovers:
        dimension_identifier = transform_spillover.get_namespaced_attr_or_none(
            "data-type", "parameter"
        )

        compensation_matrices = []
        matrix_rows = transform_spillover.findall("transforms:coefficient")
        for matrix_row in matrix_rows:
            matrix_dim_id = matrix_row.get_namespaced_attr_or_none(
                "data-type", "parameter"
            )
            value_str = matrix_row.get_namespaced_attr_or_none("transforms", "value")
            compensation_matrices.append(
                CompensationMatrix(
                    dimension_identifier=matrix_dim_id,
                    compensation_value=try_float_or_none(value_str),
                )
            )

        result.append(
            CompensationMatrixGroup(
                dimension_identifier=dimension_identifier,
                compensation_matrices=compensation_matrices,
            )
        )

    return result


def _extract_statistics(population: StrictXmlElement) -> list[Statistic] | None:
    """
    Extract statistics from a population element.

    Args:
        population: The population element to extract statistics from

    Returns:
        list[Statistic] | None: List of statistics if found, None otherwise
    """
    fluorescence_dimensions = []
    count_dimensions = []

    statistic_elements = population.findall("Statistic")
    if not statistic_elements:
        return None

    for statistic in statistic_elements:
        name = statistic.get_attr_or_none("name")
        dimension_id = statistic.get_attr_or_none("id")
        value_str = statistic.get_attr_or_none("value")

        if not (name and value_str):
            continue

        value = try_float_or_none(value_str)
        if value is None or math.isnan(value):
            continue

        stat_info = FLOWJO_STATISTIC_MAP.get(name)
        if stat_info is None:
            continue

        dimension = StatisticDimension(
            dimension_identifier=dimension_id,
            value=value * 100 if name in COUNT_FIELDS else value,
            unit=stat_info["unit"],
            has_statistic_datum_role=stat_info["role"],
        )

        if name in FLUORESCENCE_FIELDS:
            fluorescence_dimensions.append(dimension)
        elif name in COUNT_FIELDS:
            count_dimensions.append(dimension)

    statistics = []

    if fluorescence_dimensions:
        statistics.append(
            Statistic(
                statistical_feature="fluorescence",
                statistic_dimension=fluorescence_dimensions,
            )
        )

    if count_dimensions:
        statistics.append(
            Statistic(statistical_feature="count", statistic_dimension=count_dimensions)
        )

    return statistics if statistics else None


def _create_populations(
    node: StrictXmlElement, parent_id: str | None = None
) -> list[Population]:
    populations = []

    for population in node.findall("Population"):
        written_name = population.get_attr_or_none("name")
        count = population.get_attr_or_none("count")
        current_id = random_uuid_str()
        gate = population.find_or_none("Gate")
        group_identifier = population.get_attr_or_none("owningGroup")
        data_region_identifier = None
        if gate is not None:
            data_region_identifier = gate.get_namespaced_attr_or_none("gating", "id")

        statistics = None
        subpops_element = population.find_or_none("Subpopulations")
        if subpops_element is not None:
            statistics = _extract_statistics(subpops_element)
            sub_populations = _create_populations(subpops_element, current_id)
        else:
            sub_populations = []

        pop = Population(
            population_identifier=current_id,
            parent_population_identifier=parent_id,
            written_name=written_name,
            data_region_identifier=data_region_identifier,
            count=int(count) if count else None,
            sub_populations=sub_populations,
            custom_info={"group_identifier": group_identifier},
            statistics=statistics,
        )
        populations.append(pop)

    return populations


def _process_sample(sample: StrictXmlElement) -> list[Population]:
    sample_node = sample.find_or_none("SampleNode")
    if sample_node is None:
        return []

    subpops_element = sample_node.find_or_none("Subpopulations")
    root_id = random_uuid_str()
    sub_populations = []
    if subpops_element is not None:
        sub_populations = _create_populations(subpops_element, root_id)

    count_str = sample_node.get_attr_or_none("count")
    root_population = Population(
        population_identifier=root_id,
        parent_population_identifier=None,
        written_name=sample_node.get_attr_or_none("name"),
        data_region_identifier=None,
        count=int(count_str) if count_str else None,
        sub_populations=sub_populations,
        statistics=None,
    )

    return [root_population]


def _extract_dimension_identifiers(
    gate_element: StrictXmlElement,
) -> tuple[str | None, str | None]:
    """
    Extract dimension identifiers from a gate element.

    Args:
        gate_element: The gate element to extract dimensions from

    Returns:
        tuple[str | None, str | None]: A tuple containing the x and y dimension identifiers
    """
    x_dim = None
    y_dim = None

    dimensions = gate_element.findall("gating:dimension")

    # Extract x dimension (first dimension)
    if len(dimensions) > 0:
        x_dimension = dimensions[0]
        fcs_dimension = x_dimension.find_or_none("data-type:fcs-dimension")
        if fcs_dimension is not None:
            x_dim = fcs_dimension.get_namespaced_attr_or_none("data-type", "name")

    # Extract y dimension (second dimension)
    if len(dimensions) > 1:
        y_dimension = dimensions[1]
        fcs_dimension = y_dimension.find_or_none("data-type:fcs-dimension")
        if fcs_dimension is not None:
            y_dim = fcs_dimension.get_namespaced_attr_or_none("data-type", "name")

    return x_dim, y_dim


def _get_gate_type(gate_element: StrictXmlElement) -> str | None:
    # Handle special case for CurlyQuad
    if gate_element.find_or_none("gating:CurlyQuad") is not None:
        return RegionType.CURLY_QUAD.value

    for region_type in RegionType:
        if gate_element.find_or_none(f"gating:{region_type.value}Gate") is not None:
            return region_type.value
    return None


def _extract_vertices(
    gate_element: StrictXmlElement,
    gate_type: str | None,
    x_dim: str | None = None,
    y_dim: str | None = None,
) -> list[Vertex] | None:
    """
    Extract vertex coordinates from a gate element.

    Handles different gate types:
    - Polygon: Extract vertices from gating:vertex elements
    - Rectangle: Extract min/max coordinates to create 4 vertices
    - CurlyQuad: Extract min/max coordinates to create a vertex
    - Ellipsoid: Extract foci (2 vertices) and edge (4 vertices) coordinates

    Returns:
        list[Vertex] | None: List of vertices if found, None otherwise
    """
    extractor = VertexExtractor.build(gate_element, gate_type, x_dim, y_dim)
    if extractor is None:
        return None
    return extractor.extract()


def _create_data_regions(sample: StrictXmlElement) -> list[DataRegion]:
    data_regions: list[DataRegion] = []
    sample_node = sample.find_or_none("SampleNode")

    def _process_population(population: StrictXmlElement) -> None:
        if not (gate := population.find_or_none("Gate")) or not (
            gate_type := _get_gate_type(gate)
        ):
            return
        if gate_type == RegionType.CURLY_QUAD.value:
            gate_element = gate.find_or_none("gating:CurlyQuad")
        else:
            gate_element = gate.find_or_none(f"gating:{gate_type}Gate")

        if not gate_element:
            return

        x_dim, y_dim = _extract_dimension_identifiers(gate_element)
        vertices = _extract_vertices(gate_element, gate_type, x_dim, y_dim)
        data_regions.append(
            DataRegion(
                region_data_identifier=gate.get_namespaced_attr_or_none("gating", "id"),
                region_data_type=gate_type,
                parent_data_region_identifier=gate.get_namespaced_attr_or_none(
                    "gating", "parent_id"
                ),
                x_coordinate_dimension_identifier=x_dim,
                y_coordinate_dimension_identifier=y_dim,
                vertices=vertices,
            )
        )

    def process_population(population: StrictXmlElement | None) -> None:
        if not population:
            return
        _process_population(population)
        if not (subpops_element := population.find_or_none("Subpopulations")):
            return
        for subpop in subpops_element.findall("Population"):
            process_population(subpop)

    process_population(sample_node)

    return data_regions


def create_measurement_groups(root_element: StrictXmlElement) -> list[MeasurementGroup]:
    sample_list = root_element.find_or_none("SampleList")
    if sample_list is None:
        msg = "No SampleList element found in XML file."
        raise AllotropeParsingError(msg)

    samples = sample_list.findall("Sample")

    result = []
    for sample in samples:
        sample_node = sample.find("SampleNode")
        experimental_data_identifier = sample_node.get_attr_or_none("name")

        measurement_group = MeasurementGroup(
            experimental_data_identifier=experimental_data_identifier,
            measurement_time=_get_measurement_time(sample),
            analyst=_get_keyword_value_by_name_from_sample(sample, "$OP"),
            compensation_matrix_groups=_create_compensation_matrix_groups(sample),
            measurements=[
                Measurement(
                    measurement_identifier=random_uuid_str(),
                    sample_identifier=sample_node.get_attr("sampleID"),
                    location_identifier=_get_keyword_value_by_name_from_sample(
                        sample, "WELL ID"
                    ),
                    well_plate_identifier=_get_keyword_value_by_name_from_sample(
                        sample, "PLATE ID"
                    ),
                    written_name=_get_keyword_value_by_name_from_sample(sample, "$SRC"),
                    device_type=constants.DEVICE_TYPE,
                    method_version=root_element.get_attr_or_none("version"),
                    data_processing_time=root_element.get_attr_or_none("modDate"),
                    processed_data_identifier=random_uuid_str(),
                    populations=_process_sample(sample),
                    data_regions=_create_data_regions(sample),
                )
            ],
        )
        result.append(measurement_group)

    return result


def _get_measurement_time(sample: StrictXmlElement) -> str | None:
    date = _get_keyword_value_by_name_from_sample(sample, "$DATE")
    etim = _get_keyword_value_by_name_from_sample(sample, "$ETIM")
    if date is None or etim is None:
        return None
    return f"{date} {etim}"



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass, field

from allotropy.calcdocs.extractor import Element
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.utils.values import assert_not_none


@dataclass(frozen=True)
class Key:
    name: str
    value: str


@dataclass(frozen=True)
class Keys:
    entries: tuple[Key, ...] = field(default_factory=tuple)

    def is_empty(self) -> bool:
        return len(self.entries) == 0

    def add(self, name: str, value: str) -> Keys:
        return Keys(entries=(*self.entries, Key(name, value)))

    def insert(self, name: str, value: str, index: int = 0) -> Keys:
        data = list(self.entries)
        data.insert(index, Key(name, value))
        return Keys(entries=tuple(data))

    def append(self, keys: Keys) -> Keys:
        return Keys(self.entries + keys.entries)

    def get_or_none(self, name: str) -> Key | None:
        for key in self.entries:
            if key.name == name:
                return key
        return None

    def get(self, name: str) -> Key:
        return assert_not_none(
            self.get_or_none(name),
            msg=f"Unable to find key '{name}'",
        )

    def get_idx_or_none(self, name: str) -> int | None:
        for idx, key in enumerate(self.entries):
            if key.name == name:
                return idx
        return None

    def get_idx(self, name: str) -> int:
        return assert_not_none(
            self.get_idx_or_none(name),
            msg=f"Unable to find index of key '{name}'",
        )

    def overwrite(self, name: str, value: str) -> Keys:
        index = self.get_idx_or_none(name)
        if index is None:
            return self.insert(name, value, index=0)

        return Keys(
            entries=tuple(
                key if idx != index else Key(name, value)
                for idx, key in enumerate(self.entries)
            )
        )

    def delete(self, name: str) -> Keys:
        return Keys(entries=tuple(key for key in self.entries if key.name != name))

    def extract(self, name: str) -> tuple[Key, Keys]:
        return self.get(name), self.delete(name)


class ViewData:
    def __init__(
        self,
        view: View,
        name: str,
        data: dict[str, ViewData | list[Element]],
    ):
        self.view = view
        self.name = name
        self.data = data

    def filter_keys(self, keys: Keys) -> Keys:
        return self.view.filter_keys(keys)

    def iter_keys(self) -> Iterator[Keys]:
        for data_key, item in self.data.items():
            keys = Keys().add(self.name, data_key)
            if isinstance(item, ViewData):
                for sub_keys in item.iter_keys():
                    yield keys.append(sub_keys)
            else:
                yield keys

    def get_item(self, keys: Keys) -> ViewData | list[Element]:
        if keys.is_empty():
            return self

        key, new_keys = keys.extract(self.name)
        item = self.data[key.value]
        return item.get_item(new_keys) if isinstance(item, ViewData) else item

    def get_sub_view_data(self, keys: Keys) -> ViewData:
        item = self.get_item(keys)
        if isinstance(item, ViewData):
            return item
        msg = f"Unable to find sub view data with keys: {keys}."
        raise AllotropyParserError(msg)

    def get_leaf_items(self, keys: Keys) -> list[Element]:
        item = self.get_item(keys)
        if isinstance(item, ViewData):
            msg = f"Unable to find leaf item of view data with keys: {keys}."
            raise AllotropyParserError(msg)
        return item


class View(ABC):
    def __init__(self, name: str, sub_view: View | None = None):
        self.name = name
        self.sub_view = sub_view

    @abstractmethod
    def sort_elements(self, _: list[Element]) -> dict[str, list[Element]]:
        pass

    def filter_keys(self, keys: Keys) -> Keys:
        filtered_keys = self.sub_view.filter_keys(keys) if self.sub_view else Keys()
        if key := keys.get_or_none(self.name):
            return filtered_keys.overwrite(self.name, key.value)
        return filtered_keys

    def apply(self, elements: list[Element]) -> ViewData:
        return ViewData(
            view=self,
            name=self.name,
            data={
                id_: self.sub_view.apply(element) if self.sub_view else element
                for id_, element in self.sort_elements(elements).items()
            },
        )



from __future__ import annotations

from xml.etree import ElementTree

# xml fromstring is vulnerable so defusedxml version is used instead
from defusedxml.ElementTree import fromstring

from allotropy.parsers.utils.values import assert_not_none, try_float_or_none


class StrictXmlElement:
    @classmethod
    def create_from_bytes(cls, data: bytes) -> StrictXmlElement:
        return StrictXmlElement(fromstring(data))

    def __init__(
        self, element: ElementTree.Element, namespaces: dict[str, str] | None = None
    ):
        self.element = element
        self.namespaces = namespaces or {}

    def find_or_none(self, name: str) -> StrictXmlElement | None:
        element = self.element.find(name, self.namespaces)
        return (
            StrictXmlElement(element, self.namespaces) if element is not None else None
        )

    def find(self, name: str) -> StrictXmlElement:
        return assert_not_none(
            self.find_or_none(name),
            msg=f"Unable to find '{name}' in xml file contents",
        )

    def recursive_find_or_none(self, names: list[str]) -> StrictXmlElement | None:
        if len(names) == 0:
            return self
        name, *sub_names = names
        if element := self.find_or_none(name):
            return element.recursive_find_or_none(sub_names)
        return None

    def recursive_find(self, names: list[str]) -> StrictXmlElement:
        if len(names) == 0:
            return self
        name, *sub_names = names
        return self.find(name).recursive_find(sub_names)

    def findall(self, name: str) -> list[StrictXmlElement]:
        return [
            StrictXmlElement(element, self.namespaces)
            for element in self.element.findall(name, self.namespaces)
        ]

    def get_attr_or_none(self, name: str) -> str | None:
        value = self.element.get(name)
        return None if value is None else str(value)

    def get_attr(self, name: str) -> str:
        return assert_not_none(
            self.get_attr_or_none(name),
            msg=f"Unable to find '{name}' in xml file contents",
        )

    def parse_text_or_none(self) -> StrictXmlElement | None:
        if (text := self.get_text_or_none()) is None:
            return None
        try:
            return StrictXmlElement(fromstring(text), self.namespaces)
        except ElementTree.ParseError:
            return None

    def parse_text(self, name: str) -> StrictXmlElement:
        return assert_not_none(
            self.parse_text_or_none(),
            msg=f"Unable to parse text from xml tag '{name}' as valid xml content",
        )

    def get_text_or_none(self) -> str | None:
        return self.element.text

    def get_text(self, name: str) -> str:
        return assert_not_none(
            self.get_text_or_none(),
            msg=f"Unable to find valid string from xml tag '{name}'",
        )

    def get_float_or_none(self) -> float | None:
        return try_float_or_none(self.get_text_or_none())

    def get_float(self, name: str) -> float:
        return assert_not_none(self.get_float_or_none(), name)

    def get_sub_float_or_none(self, name: str) -> float | None:
        if element := self.find_or_none(name):
            return element.get_float_or_none()
        return None

    def get_sub_text_or_none(self, name: str) -> str | None:
        if element := self.find_or_none(name):
            return element.get_text_or_none()
        return None

    # namespace-specific methods
    def get_namespaced_attr_or_none(self, namespace_key: str, field: str) -> str | None:
        if namespace_key not in self.namespaces:
            return None
        return self.element.get(f"{{{self.namespaces.get(namespace_key)}}}{field}")

    def get_namespaced_attr(self, namespace_key: str, field: str) -> str:
        return assert_not_none(
            self.get_namespaced_attr_or_none(namespace_key, field),
            msg=f"Unable to find '{namespace_key}:{field}' in xml file contents",
        )



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field

import pandas as pd

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueSecondTime,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    MeasurementType,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.cytiva_biacore_t200_control import constants
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float_or_none,
    try_int_or_none,
)
from allotropy.types import DictType


@dataclass(frozen=True)
class ChipData:
    sensor_chip_identifier: str
    sensor_chip_type: str | None
    number_of_flow_cells: int | None
    number_of_spots: int | None
    lot_number: str | None
    custom_info: DictType

    @staticmethod
    def create(chip_data: DictType) -> ChipData:
        return ChipData(
            sensor_chip_identifier=assert_not_none(chip_data.get("Id"), "Chip ID"),
            sensor_chip_type=chip_data.get("Name"),
            number_of_flow_cells=try_int_or_none(chip_data.get("NoFcs")),
            number_of_spots=try_int_or_none(chip_data.get("NoSpots")),
            lot_number=lot_no if (lot_no := chip_data.get("LotNo")) else None,
            custom_info={
                "ifc identifier": chip_data.get("IFC"),
                "last modified time": chip_data.get("LastModTime"),
                "last use time": chip_data.get("LastUseTime"),
                "first dock date": chip_data.get("FirstDockDate"),
            },
        )


@dataclass(frozen=True)
class DetectionSetting:
    key: str
    value: float

    @staticmethod
    def create(detection_setting: DictType) -> DetectionSetting:
        detection_key = f"Detection{detection_setting['Detection']}"
        return DetectionSetting(
            key=detection_key.lower(),
            value=detection_setting[detection_key],
        )


@dataclass(frozen=True)
class Device:
    type_: str
    identifier: str


@dataclass(frozen=True)
class RunMetadata:
    analyst: str | None = None
    compartment_temperature: float | None = None
    baseline_flow: float | None = None
    data_collection_rate: float | None = None
    detection_setting: DetectionSetting | None = None
    buffer_volume: float | None = None
    devices: list[Device] = field(default_factory=list)

    @staticmethod
    def create(
        application_template_details: dict[str, DictType] | None,
    ) -> RunMetadata:
        if application_template_details is None:
            return RunMetadata()
        return RunMetadata(
            analyst=application_template_details["properties"].get("User"),
            compartment_temperature=try_float_or_none(
                application_template_details["RackTemperature"].get("Value")
                if "sample_data" in application_template_details
                else application_template_details["system_preparations"].get("RackTemp")
            ),
            baseline_flow=try_float_or_none(
                application_template_details.get("BaselineFlow", {}).get("value")
            ),
            data_collection_rate=try_float_or_none(
                application_template_details.get("DataCollectionRate", {}).get("value")
            ),
            detection_setting=(
                DetectionSetting.create(detection_setting)
                if (detection_setting := application_template_details.get("detection"))
                else None
            ),
            buffer_volume=try_float_or_none(
                next(
                    value
                    for key, value in application_template_details[
                        "prepare_run"
                    ].items()
                    if key.startswith("Buffer")
                )
            ),
            devices=[
                Device(type_=key.split()[0], identifier=key.split()[-1])
                for key in application_template_details
                if key.startswith("Flowcell")
            ],
        )


@dataclass(frozen=True)
class SystemInformation:
    device_identifier: str
    model_number: str
    measurement_time: str
    experiment_type: str | None
    analytical_method_identifier: str | None
    software_name: str | None
    software_version: str | None

    @staticmethod
    def create(system_information: DictType) -> SystemInformation:
        return SystemInformation(
            device_identifier=assert_not_none(
                system_information.get("InstrumentId"), "InstrumentId"
            ),
            model_number=assert_not_none(
                system_information.get("ProcessingUnit"), "ProcessingUnit"
            ),
            measurement_time=assert_not_none(
                system_information.get("Timestamp"), "Timestamp"
            ),
            experiment_type=system_information.get("RunTypeId"),
            analytical_method_identifier=system_information.get("TemplateFile"),
            software_name=system_information.get("Application"),
            software_version=system_information.get("Version"),
        )


@dataclass(frozen=True)
class ReportPointData:
    identifier: str
    identifier_role: str
    absolute_resonance: float
    relative_resonance: float | None
    time_setting: float
    custom_info: DictType
    min_resonance: float
    max_resonance: float
    lrsd: float
    slope: float
    sd: float

    @staticmethod
    def create(data: SeriesData) -> ReportPointData:
        return ReportPointData(
            identifier=random_uuid_str(),
            identifier_role=data[str, "Id"],
            absolute_resonance=data[float, "AbsResp"],
            relative_resonance=data.get(float, "RelResp"),
            time_setting=data[float, "Time"],
            min_resonance=data[float, "Min"],
            max_resonance=data[float, "Max"],
            lrsd=data[float, "LRSD"],
            slope=data[float, "Slope"],
            sd=data[float, "SD"],
            custom_info={
                "diode_row": data.get(str, "DiodeRow"),
                "window": (
                    TQuantityValueSecondTime(value=window)
                    if (window := data.get(float, "Window")) is not None
                    else None
                ),
                "quality": data.get(str, "Quality"),
                "baseline": data.get(str, "Baseline"),
                "assay_step": data.get(str, "AssayStep"),
                "assay_step_purpose": data.get(str, "AssayStepPurpose"),
                "buffer": data.get(str, "Buffer"),
            },
        )


@dataclass(frozen=True)
class MeasurementData:
    identifier: str
    type_: MeasurementType
    device_type: str
    sample_identifier: str
    flow_cell_identifier: str
    sensorgram_data: pd.DataFrame
    report_point_data: list[ReportPointData] | None
    location_identifier: str | None
    sample_role_type: str | None
    concentration: float | None
    molecular_weight: float | None
    # for Mobilization experiments
    method_name: str | None
    ligand_identifier: str | None
    flow_path: str | None
    flow_rate: float | None
    contact_time: float | None
    dilution: float | None


@dataclass(frozen=True)
class SampleData:
    measurements: dict[str, list[MeasurementData]]

    @staticmethod
    def create(intermediate_structured_data: DictType) -> SampleData:
        application_template_details: dict[
            str, DictType
        ] = intermediate_structured_data.get("application_template_details", {})
        measurements: dict[str, list[MeasurementData]] = defaultdict(list)
        for idx in range(intermediate_structured_data["total_cycles"]):
            flowcell_cycle_data: DictType = application_template_details.get(
                f"Flowcell {idx + 1}", {}
            )
            sample_data: DictType = (
                sd[idx]
                if (sd := intermediate_structured_data.get("sample_data"))
                else {}
            )
            cycle_data: DictType = intermediate_structured_data["cycle_data"][idx]
            sensorgram_data: pd.DataFrame = cycle_data["sensorgram_data"]
            # some experiments don't have report point data for some cycles (apparently just the first one)
            report_point_data: pd.DataFrame | None = cycle_data["report_point_data"]

            sample_identifier = sample_data.get("sample_name", NOT_APPLICABLE)
            location_identifier = sample_data.get("rack")
            sample_location_key = f"{location_identifier}_{sample_identifier}"

            # Measurements are grouped by sample and location identifiers
            measurements[sample_location_key] += [
                MeasurementData(
                    identifier=random_uuid_str(),
                    type_=MeasurementType.SURFACE_PLASMON_RESONANCE,
                    device_type=constants.DEVICE_TYPE,
                    sample_identifier=sample_identifier,
                    flow_cell_identifier=str(flow_cell),
                    location_identifier=location_identifier,
                    sample_role_type=constants.SAMPLE_ROLE_TYPE.get(
                        sample_data.get("role", "__IVALID_KEY__")
                    ),
                    concentration=try_float_or_none(sample_data.get("concentration")),
                    molecular_weight=try_float_or_none(
                        sample_data.get("molecular_weight")
                    ),
                    sensorgram_data=sensorgram_df,
                    report_point_data=(
                        map_rows(
                            report_point_data[report_point_data["Fc"] == flow_cell],
                            ReportPointData.create,
                        )
                        if report_point_data is not None
                        else None
                    ),
                    # for Mobilization experiments
                    method_name=flowcell_cycle_data.get("MethodName"),
                    ligand_identifier=flowcell_cycle_data.get("Ligand"),
                    flow_path=flowcell_cycle_data.get("DetectionText"),
                    flow_rate=try_float_or_none(flowcell_cycle_data.get("Flow")),
                    contact_time=try_float_or_none(
                        flowcell_cycle_data.get("ContactTime")
                    ),
                    dilution=try_float_or_none(
                        flowcell_cycle_data.get("DilutePercent")
                    ),
                )
                # group sensorgram data by Flow Cell Number (Fc in rpoint data)
                for flow_cell, sensorgram_df in sensorgram_data.groupby(
                    "Flow Cell Number"
                )
            ]
        return SampleData(measurements)


@dataclass(frozen=True)
class Data:
    run_metadata: RunMetadata
    chip_data: ChipData
    system_information: SystemInformation
    sample_data: SampleData

    @staticmethod
    def create(intermediate_structured_data: DictType) -> Data:
        application_template_details: dict[
            str, DictType
        ] | None = intermediate_structured_data.get("application_template_details")
        chip_data: DictType = intermediate_structured_data["chip"]
        system_information: DictType = intermediate_structured_data[
            "system_information"
        ]
        return Data(
            run_metadata=RunMetadata.create(application_template_details),
            chip_data=ChipData.create(chip_data),
            system_information=SystemInformation.create(system_information),
            sample_data=SampleData.create(intermediate_structured_data),
        )



import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import read_csv


class ViCellBluReader:
    SUPPORTED_EXTENSIONS = "csv"

    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        return read_csv(
            named_file_contents.contents,
            index_col=False,
            encoding=named_file_contents.encoding,
        )
